from pyrelational.strategies.classification.entropy_classification_strategy import (
    EntropyClassificationStrategy,
)
from pyrelational.strategies.classification.least_confidence_strategy import (
    LeastConfidenceStrategy,
)
from pyrelational.strategies.classification.marginal_confidence_strategy import (
    MarginalConfidenceStrategy,
)
from pyrelational.strategies.classification.ratio_confidence_strategy import (
    RatioConfidenceStrategy,
)



from abc import ABC

import torch

from pyrelational.data import GenericDataManager
from pyrelational.models import GenericModel
from pyrelational.strategies.generic_al_strategy import GenericActiveLearningStrategy


class GenericRegressionStrategy(GenericActiveLearningStrategy, ABC):
    """A base active learning strategy class for regression in which the top n indices,
    according to user-specified scoring function, are queried at each iteration"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(GenericRegressionStrategy, self).__init__(data_manager, model)
        self.scoring_fn = NotImplementedError

    def active_learning_step(self, num_annotate: int):
        self.model.train(self.l_loader, self.valid_loader)
        output = self.model(self.u_loader)
        scores = self.scoring_fn(x=output)
        ixs = torch.argsort(scores, descending=True).tolist()
        return [self.u_indices[i] for i in ixs[:num_annotate]]



from pyrelational.models.ensemble_model import (
    GenericEnsembleModel,
    LightningEnsembleModel,
)
from pyrelational.models.generic_model import GenericModel
from pyrelational.models.lightning_model import LightningModel
from pyrelational.models.mcdropout_model import (
    GenericMCDropoutModel,
    LightningMCDropoutModel,
)



"""Classification datasets that can be used for benchmarking AL strategies
"""

import os
import urllib.request
from os import path

import numpy as np
import pyreadr
import scipy.io
import torch
import torch.distributions as distributions
from sklearn.datasets import load_breast_cancer, load_digits
from sklearn.model_selection import KFold, StratifiedKFold
from torch.utils.data import Dataset
from torchvision import datasets, transforms

from .uci_datasets import UCIDatasets


class SynthClass1(Dataset):
    """
    Synth1 dataset as described in Yang and Loog

    Consists of a binary classification task of positive
    and negative class samples being generated by a multivariate
    gaussian distribution centered at [1,1] and [-1,-1]
    respectively.

    :param n_splits: an int describing the number of class stratified
            splits to compute
    :param size: an int describing the number of observations the dataset
            is to have
    :param random_seed: random seed for reproducibility on splits
    """

    def __init__(self, n_splits=5, size=500, random_seed=1234):
        super(SynthClass1, self).__init__()
        self.n_splits = n_splits
        pos_distribution = distributions.MultivariateNormal(torch.FloatTensor([3, 3]), torch.eye(2))
        neg_distribution = distributions.MultivariateNormal(torch.FloatTensor([0, 0]), torch.eye(2))

        num_pos = int(size / 2.0)
        num_neg = size - num_pos
        pos_samples = torch.vstack([pos_distribution.sample() for _ in range(num_pos)])
        neg_samples = torch.vstack([neg_distribution.sample() for _ in range(num_neg)])
        pos_targets = torch.ones(num_pos, dtype=torch.long)
        neg_targets = torch.ones(num_neg, dtype=torch.long) * 0

        self.x = torch.cat([pos_samples, neg_samples])
        self.y = torch.cat([pos_targets, neg_targets])

        skf = StratifiedKFold(n_splits=n_splits)
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class SynthClass2(Dataset):
    """
    Synth2 dataset as described in Yang and Loog

    Originally proposed by Huang et al in:
    Active learning by querying informative
    and representative examples

    :param n_splits: an int describing the number of class stratified
            splits to compute
    :param size: an int describing the number of observations the dataset
            is to have
    :param random_seed: random seed for reproducibility on splits
    """

    def __init__(self, n_splits=5, size=500, random_seed=1234):
        super(SynthClass2, self).__init__()
        self.n_splits = n_splits

        pos_dist_1 = distributions.MultivariateNormal(torch.FloatTensor([0, 5]), torch.eye(2))
        neg_dist_1 = distributions.MultivariateNormal(torch.FloatTensor([0, -5]), torch.eye(2))

        pos_dist_2 = distributions.MultivariateNormal(torch.FloatTensor([-5, 10]), torch.eye(2))
        pos_dist_3 = distributions.MultivariateNormal(torch.FloatTensor([-5, -10]), torch.eye(2))

        neg_dist_2 = distributions.MultivariateNormal(torch.FloatTensor([5, 10]), torch.eye(2))
        neg_dist_3 = distributions.MultivariateNormal(torch.FloatTensor([5, -10]), torch.eye(2))

        num_pos = int(size / 2.0)
        num_neg = size - num_pos

        # find number of samples to generate from the positives and negative blobs constrained
        # to input size
        num_pos1, num_pos2, num_pos3 = [len(x) for x in self._split(range(num_pos), 3)]
        num_neg1, num_neg2, num_neg3 = [len(x) for x in self._split(range(num_neg), 3)]

        pos_samples_1 = torch.vstack([pos_dist_1.sample() for _ in range(num_pos1)])
        pos_samples_2 = torch.vstack([pos_dist_2.sample() for _ in range(num_pos2)])
        pos_samples_3 = torch.vstack([pos_dist_3.sample() for _ in range(num_pos3)])

        neg_samples_1 = torch.vstack([neg_dist_1.sample() for _ in range(num_neg1)])
        neg_samples_2 = torch.vstack([neg_dist_2.sample() for _ in range(num_neg2)])
        neg_samples_3 = torch.vstack([neg_dist_3.sample() for _ in range(num_neg3)])

        pos_targets = torch.ones(num_pos, dtype=torch.long)
        neg_targets = torch.ones(num_neg, dtype=torch.long) * 0

        self.x = torch.cat([pos_samples_1, pos_samples_2, pos_samples_3, neg_samples_1, neg_samples_2, neg_samples_3])
        self.y = torch.cat([pos_targets, neg_targets])

        skf = StratifiedKFold(n_splits=n_splits)
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def _split(self, iterable, n):
        # split the iterable into n approximately same size parts
        k, m = divmod(len(iterable), n)
        return (iterable[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class SynthClass3(Dataset):
    """SynthClass3 dataset as described in Yang and Loog

    :param n_splits: an int describing the number of class stratified
            splits to compute
    :param size: an int describing the number of observations the dataset
            is to have
    :param random_seed: random seed for reproducibility on splits
    """

    def __init__(self, n_splits=5, size=500, random_seed=1234):
        super(SynthClass3, self).__init__()
        self.size = size
        self.random_seed = random_seed
        self.n_splits = n_splits

        cov = torch.FloatTensor([[0.60834549, -0.63667341], [-0.40887718, 0.85253229]])
        cov = torch.matmul(cov, cov.T)

        pos_dist_1 = distributions.MultivariateNormal(torch.FloatTensor([0, 0]), cov)
        pos_dist_2 = distributions.MultivariateNormal(torch.FloatTensor([3, 10]), cov)
        neg_dist_1 = distributions.MultivariateNormal(torch.FloatTensor([3, 3]), torch.FloatTensor([[1, 2], [2, 7]]))

        num_pos = int(size / 2.0)
        num_neg1 = size - num_pos
        num_pos1, num_pos2 = [len(x) for x in self._split(range(num_pos), 2)]

        pos_samples_1 = torch.vstack([pos_dist_1.sample() for _ in range(num_pos1)])
        pos_samples_2 = torch.vstack([pos_dist_2.sample() for _ in range(num_pos2)])
        neg_samples_1 = torch.vstack([neg_dist_1.sample() for _ in range(num_neg1)])

        pos_targets = torch.ones(num_pos, dtype=torch.long)
        neg_targets = torch.ones(num_neg1, dtype=torch.long) * 0

        self.x = torch.cat([pos_samples_1, pos_samples_2, neg_samples_1])
        self.y = torch.cat([pos_targets, neg_targets])

        skf = StratifiedKFold(n_splits=n_splits)
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def _split(self, iterable, n):
        # split the iterable into n approximately same size parts
        k, m = divmod(len(iterable), n)
        return (iterable[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class BreastCancerDataset(Dataset):
    """UCI ML Breast Cancer Wisconsin (Diagnostic) dataset

    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, n_splits=5):
        super(BreastCancerDataset, self).__init__()
        sk_x, sk_y = load_breast_cancer(return_X_y=True)
        self.x = torch.FloatTensor(sk_x)
        self.y = torch.LongTensor(sk_y)

        skf = StratifiedKFold(n_splits=n_splits)
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class DigitDataset(Dataset):
    """UCI ML hand-written digits datasets

    From C. Kaynak (1995) Methods of Combining Multiple Classifiers and
    Their Applications to Handwritten Digit Recognition, MSc Thesis,
    Institute of Graduate Studies in Science and Engineering, Bogazici
    University.

    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, n_splits=5):
        super(DigitDataset, self).__init__()
        sk_x, sk_y = load_digits(return_X_y=True)
        self.x = torch.FloatTensor(sk_x)  # data
        self.y = torch.LongTensor(sk_y)  # target

        skf = StratifiedKFold(n_splits=n_splits)
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class FashionMNIST(Dataset):
    """Fashion MNIST Dataset

    From Fashion-MNIST: a Novel Image Dataset for Benchmarking Machine Learning
    Algorithms. Han Xiao, Kashif Rasul, Roland Vollgraf. arXiv:1708.07747

    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(FashionMNIST, self).__init__()
        train_dataset = datasets.FashionMNIST(root=data_dir, train=True, download=True, transform=transforms.ToTensor())
        test_dataset = datasets.FashionMNIST(root=data_dir, train=False, download=True, transform=transforms.ToTensor())
        dataset = torch.utils.data.ConcatDataset([train_dataset, test_dataset])
        self.x = torch.stack([(dataset[i][0]).flatten() for i in range(len(dataset))])
        self.y = torch.stack([torch.tensor(dataset[i][1]) for i in range(len(dataset))])

        skf = StratifiedKFold(n_splits=n_splits)
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class UCIClassification(Dataset):
    """UCI classification abstract class

    :param name: string denotation for dataset to download
        as specified in uci_datasets.UCIDatasets
    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, name, data_dir="/tmp/", n_splits=5):
        super(UCIClassification, self).__init__()
        dataset = UCIDatasets(name=name, data_dir=data_dir, n_splits=n_splits)
        torch_dataset = dataset.get_simple_dataset()

        self.data_dir = dataset.data_dir
        self.name = dataset.name
        self.data_splits = dataset.data_splits

        self.len_dataset = len(torch_dataset)
        self.x = torch_dataset[:][0]
        self.y = torch_dataset[:][1].squeeze()

    def __len__(self):
        return self.len_dataset

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


def remap_to_int(torch_class_array):
    """Remaps the values in the torch_class_array to integers from 0
    to n for n unique values in the torch_class_array

    :param torch_class_array: class array whose elements are to be
        mapped to contiguous ints
    """
    remapped_array = []
    tca2idx = {}
    mapping_value = 0
    for val in torch_class_array:
        val = int(val)
        if val in tca2idx.keys():
            remapped_array.append(tca2idx[val])
        else:
            tca2idx[val] = mapping_value
            mapping_value += 1
            remapped_array.append(tca2idx[val])
    return torch.Tensor(remapped_array)


class UCIGlass(UCIClassification):
    """UCI Glass dataset

    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIGlass, self).__init__(name="glass", data_dir=data_dir, n_splits=n_splits)
        self.y -= 1  # for 0 - k-1 class relabelling
        self.y = remap_to_int(self.y).long()  # UCIGlass has mislabelling


class UCIParkinsons(UCIClassification):
    """UCI Parkinsons dataset

    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIParkinsons, self).__init__(name="parkinsons", data_dir=data_dir, n_splits=n_splits)


class UCISeeds(UCIClassification):
    """UCI Seeds dataset

    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCISeeds, self).__init__(name="seeds", data_dir=data_dir, n_splits=n_splits)
        self.y -= 1  # for 0 - k-1 class relabeling


class StriatumDataset(Dataset):
    """Striatum dataset as used in Konyushkova et al. 2017

    From Ksenia Konyushkova, Raphael Sznitman, Pascal Fua 'Learning Active
    Learning from Data', NIPS 2017

    :param data_dir: path where to save the raw data default to /tmp/
    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(StriatumDataset, self).__init__()
        self.data_dir = data_dir
        self.n_splits = n_splits
        self.train_feat_url = (
            "https://github.com/ksenia-konyushkova/LAL/raw/master/data/striatum_train_features_mini.mat"
        )
        self.test_feat_url = "https://github.com/ksenia-konyushkova/LAL/raw/master/data/striatum_test_features_mini.mat"
        self.train_label_url = (
            "https://github.com/ksenia-konyushkova/LAL/raw/master/data/striatum_train_labels_mini.mat"
        )
        self.test_label_url = "https://github.com/ksenia-konyushkova/LAL/raw/master/data/striatum_test_labels_mini.mat"

        self._load_dataset()

    def _download_dataset(self, url):
        if not path.exists(self.data_dir):
            os.mkdir(self.data_dir)

        file_name = url.split("/")[-1]
        if not path.exists(self.data_dir + file_name):
            urllib.request.urlretrieve(url, self.data_dir + file_name)

    def _load_dataset(self):
        """Download, process, and get stratified splits"""

        # download
        self._download_dataset(self.train_feat_url)
        self._download_dataset(self.test_feat_url)
        self._download_dataset(self.train_label_url)
        self._download_dataset(self.test_label_url)

        # process
        train_feat = (scipy.io.loadmat(self.data_dir + "striatum_train_features_mini.mat"))["features"]
        test_feat = scipy.io.loadmat(self.data_dir + "striatum_test_features_mini.mat")["features"]
        train_label = scipy.io.loadmat(self.data_dir + "striatum_train_labels_mini.mat")["labels"]
        test_label = scipy.io.loadmat(self.data_dir + "striatum_test_labels_mini.mat")["labels"]

        self.x = np.vstack([train_feat, test_feat])
        self.y = np.vstack([train_label, test_label])

        skf = StratifiedKFold(n_splits=self.n_splits)
        self.in_dim = self.x.shape[1]
        self.out_dim = 1
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

        self.x = torch.from_numpy(self.x).float()
        self.y = torch.from_numpy(self.y).long().squeeze()
        self.y = remap_to_int(self.y).long()

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class GaussianCloudsDataset(Dataset):
    """GaussianClouds from Konyushkova et al. 2017 basically a imbalanced
    binary classification task created from multivariate gaussian blobs

    From Ksenia Konyushkova, Raphael Sznitman, Pascal Fua 'Learning Active
    Learning from Data', NIPS 2017

    :param data_dir: path where to save the raw data default to /tmp/
    :param n_splits: an int describing the number of class stratified
            splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        self.data_dir = data_dir
        self.n_splits = n_splits
        self._load_dataset()

    def _load_dataset(self, size=1000, n_dim=2, random_balance=False, n_splits=10):
        if random_balance:
            # proportion of class 1 to vary from 10% to 90%
            cl1_prop = np.random.rand()
            cl1_prop = (cl1_prop - 0.5) * 0.8 + 0.5
        else:
            cl1_prop = 0.8

        trainSize1 = int(size * cl1_prop)
        trainSize2 = size - trainSize1
        testSize1 = trainSize1 * 10
        testSize2 = trainSize2 * 10

        # Generate parameters of datasets
        mean1 = scipy.random.rand(n_dim)
        cov1 = scipy.random.rand(n_dim, n_dim) - 0.5
        cov1 = np.dot(cov1, cov1.transpose())
        mean2 = scipy.random.rand(n_dim)
        cov2 = scipy.random.rand(n_dim, n_dim) - 0.5
        cov2 = np.dot(cov2, cov2.transpose())

        # Training data generation
        trainX1 = np.random.multivariate_normal(mean1, cov1, trainSize1)
        trainY1 = np.ones((trainSize1, 1))
        trainX2 = np.random.multivariate_normal(mean2, cov2, trainSize2)
        trainY2 = np.zeros((trainSize2, 1))

        # Testing data generation
        testX1 = np.random.multivariate_normal(mean1, cov1, testSize1)
        testY1 = np.ones((testSize1, 1))
        testX2 = np.random.multivariate_normal(mean2, cov2, testSize2)
        testY2 = np.zeros((testSize2, 1))

        train_data = np.concatenate((trainX1, trainX2), axis=0)
        train_labels = np.concatenate((trainY1, trainY2))
        test_data = np.concatenate((testX1, testX2), axis=0)
        test_labels = np.concatenate((testY1, testY2))

        self.x = np.vstack([train_data, test_data])
        self.y = np.vstack([train_labels, test_labels]).squeeze()

        skf = StratifiedKFold(n_splits=self.n_splits)  # change to Stratified later
        self.in_dim = self.x.shape[1]
        self.out_dim = 1
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

        self.x = torch.from_numpy(self.x).float()
        self.y = torch.from_numpy(self.y).long().squeeze()

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class Checkerboard2x2Dataset(Dataset):
    """Checkerboard2x2 dataset from Konyushkova et al. 2017

    From Ksenia Konyushkova, Raphael Sznitman, Pascal Fua 'Learning Active
    Learning from Data', NIPS 2017

    :param data_dir: path where to save the raw data default to /tmp/
    :param n_splits: an int describing the number of class stratified
            splits to compute

    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(Checkerboard2x2Dataset, self).__init__()
        self.data_dir = data_dir
        self.n_splits = n_splits

        self.raw_train_url = "https://github.com/ksenia-konyushkova/LAL/raw/master/data/checkerboard2x2_train.npz"
        self.raw_test_url = "https://github.com/ksenia-konyushkova/LAL/raw/master/data/checkerboard2x2_test.npz"

        self._load_dataset()

    def _download_dataset(self, url):
        if not path.exists(self.data_dir):
            os.mkdir(self.data_dir)

        file_name = url.split("/")[-1]
        if not path.exists(self.data_dir + file_name):
            urllib.request.urlretrieve(url, self.data_dir + file_name)

    def _load_dataset(self):
        """Download, process, and get stratified splits"""

        # download
        self._download_dataset(self.raw_train_url)
        self._download_dataset(self.raw_test_url)

        # process
        train = np.load(self.data_dir + "checkerboard2x2_train.npz")
        test = np.load(self.data_dir + "checkerboard2x2_test.npz")

        train_feat, train_label = train["x"], train["y"]
        test_feat, test_label = test["x"], test["y"]

        self.x = np.vstack([train_feat, test_feat])
        self.y = np.vstack([train_label, test_label])

        skf = StratifiedKFold(n_splits=self.n_splits)  # change to Stratified later
        self.in_dim = self.x.shape[1]
        self.out_dim = 1
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

        self.x = torch.from_numpy(self.x).float()
        self.y = torch.from_numpy(self.y).long().squeeze()

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class Checkerboard4x4Dataset(Dataset):
    """Checkerboard 4x4 dataset from Konyushkova et al. 2017

    From Ksenia Konyushkova, Raphael Sznitman, Pascal Fua 'Learning Active
    Learning from Data', NIPS 2017

    :param data_dir: path where to save the raw data default to /tmp/
    :param n_splits: an int describing the number of class stratified
            splits to compute

    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(Checkerboard4x4Dataset, self).__init__()
        self.data_dir = data_dir
        self.n_splits = n_splits

        self.train_url = "https://github.com/ksenia-konyushkova/LAL/raw/master/data/checkerboard4x4_train.npz"
        self.test_url = "https://github.com/ksenia-konyushkova/LAL/raw/master/data/checkerboard4x4_test.npz"

        self._load_dataset()

    def _download_dataset(self, url):
        if not path.exists(self.data_dir):
            os.mkdir(self.data_dir)

        file_name = url.split("/")[-1]
        if not path.exists(self.data_dir + file_name):
            urllib.request.urlretrieve(url, self.data_dir + file_name)

    def _load_dataset(self):
        """Download, process, and get stratified splits"""

        # download
        self._download_dataset(self.train_url)
        self._download_dataset(self.test_url)

        # process
        train = np.load(self.data_dir + "checkerboard4x4_train.npz")
        test = np.load(self.data_dir + "checkerboard4x4_test.npz")

        train_feat, train_label = train["x"], train["y"]
        test_feat, test_label = test["x"], test["y"]

        self.x = np.vstack([train_feat, test_feat])
        self.y = np.vstack([train_label, test_label])

        skf = StratifiedKFold(n_splits=self.n_splits)  # change to Stratified later
        self.in_dim = self.x.shape[1]
        self.out_dim = 1
        self.data_splits = skf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

        self.x = torch.from_numpy(self.x).float()
        self.y = torch.from_numpy(self.y).long().squeeze()

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class CreditCardDataset(Dataset):
    """Credit card fraud dataset, highly unbalanced and challenging.

    From Andrea Dal Pozzolo, Olivier Caelen, Reid A. Johnson, and Gianluca Bontempi.
    Calibrating probability with undersampling for unbalanced classification. In 2015
    IEEE Symposium Series on Computational Intelligence, pages 159–166, 2015.

    We use the original data from http://www.ulb.ac.be/di/map/adalpozz/data/creditcard.Rdata
    processed using pyreadr

    :param data_dir: path where to save the raw data default to /tmp/
    :param n_splits: an int describing the number of class stratified
            splits to compute

    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(CreditCardDataset, self).__init__()
        self.raw_url = "http://www.ulb.ac.be/di/map/adalpozz/data/creditcard.Rdata"
        self.data_dir = data_dir
        self.n_splits = n_splits

        self._load_dataset()

    def _load_dataset(self):
        if not path.exists(self.data_dir):
            os.mkdir(self.data_dir)

        url = self.raw_url
        file_name = url.split("/")[-1]
        if not path.exists(self.data_dir + file_name):
            urllib.request.urlretrieve(self.raw_url, self.data_dir + file_name)

        data = pyreadr.read_r(self.data_dir + file_name)
        data = data["creditcard"]
        data.reset_index(inplace=True)
        self.df = data
        cols = data.columns
        xcols = cols[1:-1]
        ycol = "Class"
        x = data[xcols].to_numpy()
        y = data[ycol].to_numpy()
        _, y = np.unique(y, return_inverse=True)  # map string classes to ints
        self.x = x
        self.y = y

        skf = StratifiedKFold(n_splits=self.n_splits)  # change to Stratified later
        self.in_dim = len(xcols)
        self.out_dim = 1
        self.data_splits = skf.split(x, y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

        self.x = torch.from_numpy(self.x).float()
        self.y = torch.from_numpy(self.y).long().squeeze()

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]



"""
This module contains methods for scoring samples based on model uncertainty in
regression tasks


Most of these functions are simple but giving them a name and implementation
in PyTorch is useful for defining the different active learning strategies
"""

from typing import Optional, Union

import torch
from torch.distributions import Distribution


def regression_greedy_score(
    x: Optional[Union[torch.Tensor, Distribution]] = None,
    mean: Optional[torch.Tensor] = None,
    axis: int = 0,
) -> torch.Tensor:
    """
    Implements greedy scoring that returns mean score for each sample across repeats.
    Either x or mean should be provided as input.

    :param x: 2D pytorch tensor of repeat by scores (or scores by repeat) or pytorch Distribution
    :param std: 1D pytorch tensor corresponding to the standard deviation of a model's predictions for each sample
    :param axis: index of the axis along which the repeats are
    :return: 1D pytorch tensor of scores
    """
    _, mean, _ = _check_regression_informativeness_input(x, mean=mean, axis=axis)
    return mean


def regression_least_confidence(
    x: Optional[Union[torch.Tensor, Distribution]] = None,
    std: Optional[torch.Tensor] = None,
    axis: int = 0,
) -> torch.Tensor:
    """
    Implements least confidence scoring of based on input x returns std score for each sample across repeats.
    Either x or std should be provided as input.

    :param x: 2D pytorch tensor of repeat by scores (or scores by repeat) or pytorch Distribution
    :param std: 1D pytorch tensor corresponding to the standard deviation of a model's predictions for each sample
    :param axis: index of the axis along which the repeats are
    :return: 1D pytorch tensor of scores
    """
    _, _, std = _check_regression_informativeness_input(x, std=std, axis=axis)
    return std


def regression_expected_improvement(
    x: Optional[Union[torch.Tensor, Distribution]] = None,
    mean: Optional[torch.Tensor] = None,
    std: Optional[torch.Tensor] = None,
    max_label: float = 0.0,
    axis: int = 0,
    xi: float = 0.01,
) -> torch.Tensor:
    """
    Implements expected improvement based on max_label in the currently available data
    (`reference <https://doi.org/10.1023/A:1008306431147>`__).
    Either x or mean and std should be provided as input.


    :param x: 2D pytorch tensor or pytorch Distribution
    :param mean: 1D pytorch tensor corresponding to a model's mean predictions for each sample
    :param std: 1D pytorch tensor corresponding to the standard deviation of a model's predictions for each sample
    :param max_label: max label in the labelled dataset
    :param axis: index of the axis along which the repeats are
    :param xi: 2D pytorch tensor or pytorch Distribution
    :return: 1D pytorch tensor of scores
    """
    x, mean, std = _check_regression_informativeness_input(x, mean, std, axis=axis)
    if isinstance(x, torch.Tensor):
        return torch.relu(x - max_label).mean(axis).flatten()
    else:
        Z = torch.relu(std) * (mean - max_label - xi)
        N = torch.distributions.Normal(0, 1)
        cdf, pdf = N.cdf(Z), torch.exp(N.log_prob(Z))
        return torch.relu((mean - max_label - xi) * cdf + std * pdf)


def regression_upper_confidence_bound(
    x: Optional[Union[torch.Tensor, Distribution]] = None,
    mean: Optional[torch.Tensor] = None,
    std: Optional[torch.Tensor] = None,
    kappa: float = 1,
    axis: int = 0,
) -> torch.Tensor:
    """
    Implements Upper Confidence Bound (UCB) scoring (`reference <https://doi.org/10.1023/A:1013689704352>`__)
    Either x or mean and std should be provided as input.

    :param x: 2D pytorch tensor or pytorch Distribution
    :param mean: 1D pytorch tensor corresponding to a model's mean predictions for each sample
    :param std: 1D pytorch tensor corresponding to the standard deviation of a model's predictions for each sample
    :param kappa: trade-off parameter between exploitation and exploration
    :param axis: index of the axis along which the repeats are
    :return: 1D pytorch tensor of scores
    """

    _, mean, std = _check_regression_informativeness_input(x, mean, std, axis=axis)
    return mean + kappa * std


def regression_thompson_sampling(x: torch.Tensor, axis: int = 0) -> torch.Tensor:
    """
    Implements thompson sampling scoring (`reference <https://doi.org/10.1561/2200000070>`__).

    :param x: 2D pytorch tensor
    :param axis: index of the axis along which the repeats are
    :return: 1D pytorch tensor of scores
    """
    x, _, _ = _check_regression_informativeness_input(x, axis=axis)
    other_axis = (axis - 1) % 2
    idx = torch.randint(high=x.size(axis), size=(x.size(other_axis), 1))
    return x.gather(axis, idx).flatten()


def regression_bald(x: torch.Tensor, axis: int = 0) -> torch.Tensor:
    """
    Implementation of Bayesian Active Learning by Disagreement (BALD) for regression task
    (`reference <https://arxiv.org/pdf/1112.5745.pdf>`__)

    :param x: 2D pytorch Tensor
    :param axis: index of the axis along which the repeats are
    :return: 1D pytorch tensor of scores
    """
    x, _, _ = _check_regression_informativeness_input(x, axis=axis)
    x_mean = x.mean(axis, keepdim=True)
    x = (x - x_mean) ** 2
    return torch.log(1 + x.mean(axis)) / torch.tensor(2.0)


def _check_regression_informativeness_input(x=None, mean=None, std=None, axis=0):
    if x is None and mean is None and std is None:
        raise ValueError("Not all of x, mean, and std can be None.")

    if isinstance(x, torch.Tensor):
        x = x.squeeze()
        assert x.ndim == 2, "x input should be a 2D tensor"
        return x, x.mean(axis), x.std(axis)

    if isinstance(mean, torch.Tensor) and isinstance(std, torch.Tensor):
        mean, std = mean.squeeze(), std.squeeze()
        assert mean.ndim == 1, "mean input should be a 1D tensor"
        assert std.ndim == 1, "std input should be a 1D tensor"
        return None, mean, std
    elif isinstance(mean, torch.Tensor):
        mean = mean.squeeze()
        assert mean.ndim == 1, "mean input should be a 1D tensor"
        return None, mean, None
    elif isinstance(std, torch.Tensor):
        std = std.squeeze()
        assert std.ndim == 1, "std input should be a 1D tensor"
        return None, None, std

    if isinstance(x, Distribution):
        mean, std = x.mean.squeeze(), x.stddev.squeeze()
        assert mean.ndim == 1, "distribution input should be 1D"
        return None, mean, std



"""Defines and implements a random acquisition active learning strategy.
"""
from typing import List

import numpy as np

from pyrelational.data import GenericDataManager
from pyrelational.models import GenericModel
from pyrelational.strategies.generic_al_strategy import GenericActiveLearningStrategy


class RandomAcquisitionStrategy(GenericActiveLearningStrategy):
    """Implements RandomAcquisition whereby random samples from unlabelled set are chosen at each step"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(RandomAcquisitionStrategy, self).__init__(data_manager, model)

    def active_learning_step(self, num_annotate: int) -> List[int]:
        num_annotate = min(num_annotate, len(self.u_indices))
        return np.random.choice(self.u_indices, size=num_annotate, replace=False).tolist()



"""
Class to help download and do the initial processing of dataset on the UCI database
"""

import logging
import os
import urllib.request
import zipfile
from os import path

import numpy as np
import pandas as pd
import torch
from sklearn.model_selection import KFold, StratifiedKFold


class UCIDatasets:
    def __init__(self, name, data_dir="/tmp/", n_splits=10):
        self.datasets = {
            "concrete": "https://archive.ics.uci.edu/ml/machine-learning-databa"
            + "ses/concrete/compressive/Concrete_Data.xls",
            "energy": "http://archive.ics.uci.edu/ml/machine-learning-databases/00242/ENB2012_data.xlsx",
            "power": "https://archive.ics.uci.edu/ml/machine-learning-databases/00294/CCPP.zip",
            "wine": "https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-red.csv",
            "yacht": "http://archive.ics.uci.edu/ml/machine-learning-databases/00243/yacht_hydrodynamics.data",
            "glass": "https://archive.ics.uci.edu/ml/machine-learning-databases/glass/glass.data",
            "parkinsons": "https://archive.ics.uci.edu/ml/machine-learning-databases/parkinsons/parkinsons.data",
            "seeds": "https://archive.ics.uci.edu/ml/machine-learning-databases/00236/seeds_dataset.txt",
            "airfoil": "https://archive.ics.uci.edu/ml/machine-learning-databases/00291/airfoil_self_noise.dat",
        }
        self.data_dir = data_dir
        self.name = name
        self.n_splits = n_splits

        # flag for classification/regression dataset
        if name in ["glass", "parkinsons", "seeds"]:
            self.classification = True
        else:
            self.classification = False

        self._load_dataset()

    def _load_dataset(self):
        if self.name not in self.datasets:
            raise Exception("Not part of datasets supported in PyRelationAL at the moment")
        if not path.exists(self.data_dir):
            os.mkdir(self.data_dir)
        if not path.exists(self.data_dir + "UCI"):
            os.mkdir(self.data_dir + "UCI")

        url = self.datasets[self.name]
        file_name = url.split("/")[-1]
        if not path.exists(self.data_dir + "UCI/" + file_name):
            urllib.request.urlretrieve(self.datasets[self.name], self.data_dir + "UCI/" + file_name)
        data = None

        if self.name == "housing":
            data = pd.read_csv(self.data_dir + "UCI/housing.data", header=0, delimiter=r"\s+").values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "airfoil":
            data = pd.read_csv(self.data_dir + "UCI/airfoil_self_noise.dat", header=0, delimiter=r"\s+").values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "concrete":
            data = pd.read_excel(self.data_dir + "UCI/Concrete_Data.xls", header=0).values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "energy":
            data = pd.read_excel(self.data_dir + "UCI/ENB2012_data.xlsx", header=0).values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "power":
            zipfile.ZipFile(self.data_dir + "UCI/CCPP.zip").extractall(self.data_dir + "UCI/")
            data = pd.read_excel(self.data_dir + "UCI/CCPP/Folds5x2_pp.xlsx", header=0).values
            self.data = data

        elif self.name == "wine":
            data = pd.read_csv(self.data_dir + "UCI/winequality-red.csv", header=1, delimiter=";").values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "yacht":
            data = pd.read_csv(self.data_dir + "UCI/yacht_hydrodynamics.data", header=1, delimiter=r"\s+").values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "glass":
            data = pd.read_csv(self.data_dir + "UCI/glass.data", delimiter=",").values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "parkinsons":
            data = pd.read_csv(self.data_dir + "UCI/parkinsons.data", header=0, delimiter=",", index_col=0)
            # we need to get the "status" column as the target
            columns = list(data.columns)
            columns = set(columns) - set(["status"])
            reordered_columns = list(columns)
            reordered_columns.extend(["status"])
            data = data[reordered_columns]
            data = data.values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        elif self.name == "seeds":
            data = pd.read_csv(self.data_dir + "UCI/seeds_dataset.txt", delimiter=r"\s+", engine="python").values
            self.data = data[np.random.permutation(np.arange(len(data)))]

        self.in_dim = data.shape[1] - 1
        self.out_dim = 1

        if self.classification:
            x, y = self.data[:, : self.in_dim], self.data[:, self.in_dim :]
            skf = StratifiedKFold(n_splits=self.n_splits)
            self.data_splits = skf.split(x, y)
            self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]
        else:
            kf = KFold(n_splits=self.n_splits)
            self.data_splits = kf.split(data)
            self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def get_split(self, split=-1, train=True):
        if split == -1:
            split = 0
        if 0 <= split and split < self.n_splits:
            train_index, test_index = self.data_splits[split]
            x_train, y_train = self.data[train_index, : self.in_dim], self.data[train_index, self.in_dim :]
            x_test, y_test = self.data[test_index, : self.in_dim], self.data[test_index, self.in_dim :]
            x_means, x_stds = x_train.mean(axis=0), x_train.var(axis=0) ** 0.5
            y_means, y_stds = y_train.mean(axis=0), y_train.var(axis=0) ** 0.5
            x_train = (x_train - x_means) / x_stds
            y_train = (y_train - y_means) / y_stds
            x_test = (x_test - x_means) / x_stds
            y_test = (y_test - y_means) / y_stds
            if train:
                inps = torch.from_numpy(x_train).float()
                tgts = torch.from_numpy(y_train).float()
                train_data = torch.utils.data.TensorDataset(inps, tgts)
                return train_data
            else:
                inps = torch.from_numpy(x_test).float()
                tgts = torch.from_numpy(y_test).float()
                test_data = torch.utils.data.TensorDataset(inps, tgts)
                return test_data

    def get_full_split(self, split=-1):
        """Returns a single dataset with the test observations stacked on the train observations"""
        if split == -1:
            split = 0
        if 0 <= split and split < self.n_splits:
            train_index, test_index = self.data_splits[split]
            x_train, y_train = self.data[train_index, : self.in_dim], self.data[train_index, self.in_dim :]
            x_test, y_test = self.data[test_index, : self.in_dim], self.data[test_index, self.in_dim :]

            if self.classification:
                # standardize
                x_means, x_stds = x_train.mean(axis=0), x_train.var(axis=0) ** 0.5
                x_train = (x_train - x_means) / x_stds
                x_test = (x_test - x_means) / x_stds

                inps = torch.from_numpy(x_train).float()
                tgts = torch.from_numpy(y_train).long()
                inps_test = torch.from_numpy(x_test).float()
                tgts_test = torch.from_numpy(y_test).long()
            else:
                # standardize
                x_means, x_stds = x_train.mean(axis=0), x_train.var(axis=0) ** 0.5
                y_means, y_stds = y_train.mean(axis=0), y_train.var(axis=0) ** 0.5
                x_train = (x_train - x_means) / x_stds
                y_train = (y_train - y_means) / y_stds
                x_test = (x_test - x_means) / x_stds
                y_test = (y_test - y_means) / y_stds

                inps = torch.from_numpy(x_train).float()
                tgts = torch.from_numpy(y_train).float()
                inps_test = torch.from_numpy(x_test).float()
                tgts_test = torch.from_numpy(y_test).float()

            inps = torch.cat([inps, inps_test])
            tgts = torch.cat([tgts, tgts_test])

            dataset = torch.utils.data.TensorDataset(inps, tgts)
            return dataset

    def get_simple_dataset(self):
        """Simply return the dataset so that we can apply the splits on top after"""
        x, y = self.data[:, : self.in_dim], self.data[:, self.in_dim :]

        if self.classification:
            inps = torch.from_numpy(x).float()
            tgts = torch.from_numpy(y).long()
        else:
            inps = torch.from_numpy(x).float()
            tgts = torch.from_numpy(y).float()

        dataset = torch.utils.data.TensorDataset(inps, tgts)
        return dataset



import logging
import random
from typing import Any, Callable, List, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset, Sampler, Subset

T = TypeVar("T")
logger = logging.getLogger()


class GenericDataManager(object):
    def __init__(
        self,
        dataset: Dataset,
        train_indices: Optional[List[int]] = None,
        labelled_indices: Optional[List[int]] = None,
        unlabelled_indices: Optional[List[int]] = None,
        validation_indices: Optional[List[int]] = None,
        test_indices: Optional[List[int]] = None,
        random_label_size: Union[float, int] = 0.1,
        hit_ratio_at: Optional[Union[int, float]] = None,
        random_seed: int = 1234,
        loader_batch_size: Union[int, str] = 1,
        loader_shuffle: bool = False,
        loader_sampler: Optional[Sampler[int]] = None,
        loader_batch_sampler: Optional[Sampler[Sequence[int]]] = None,
        loader_num_workers: int = 0,
        loader_collate_fn: Optional[Callable[[List[T]], Any]] = None,
        loader_pin_memory: bool = False,
        loader_drop_last: bool = False,
        loader_timeout: float = 0,
    ):
        """
        DataManager for active learning pipelines

        :param dataset: A PyTorch dataset whose indices refer to individual samples of study
        :param train_indices: An iterable of indices mapping to training sample indices in the dataset
        :param labelled_indices: An iterable of indices  mapping to labelled training samples
        :param unlabelled_indices: An iterable of indices to unlabelled observations in the dataset
        :param validation_indices: An iterable of indices to observations used for model validation
        :param test_indices: An iterable of indices to observations in the input dataset used for
        test performance of the model
        :param random_label_size: Only used when labelled and unlabelled indices are not provided. Sets the size of
        labelled set (should either be the number of samples or ratio w.r.t. train set)
        :param hit_ratio_at: optional argument setting the top percentage threshold to compute hit ratio metric
        :param random_seed: random seed
        :param loader_batch_size: batch size for dataloader
        :param loader_shuffle: shuffle flag for dataloader
        :param loader_sampler: a sampler for the dataloaders
        :param loader_batch_sampler: a batch sampler for the dataloaders
        :param loader_num_workers: number of cpu workers for dataloaders
        :param loader_collate_fn: collate fn for dataloaders
        :param loader_pin_memory: pin memory flag for dataloaders
        :param loader_drop_last: drop last flag for dataloaders
        :param loader_timeout: timeout value for dataloaders
        """
        super(GenericDataManager, self).__init__()
        self.dataset = dataset
        self.train_indices = train_indices
        self.labelled_indices = labelled_indices
        self.unlabelled_indices = unlabelled_indices
        self.validation_indices = validation_indices
        self.test_indices = test_indices

        # Loader specific arguments
        self.loader_batch_size = loader_batch_size
        self.loader_shuffle = loader_shuffle
        self.loader_sampler = loader_sampler
        self.loader_batch_sampler = loader_batch_sampler
        self.loader_num_workers = loader_num_workers
        self.loader_collate_fn = loader_collate_fn
        self.loader_pin_memory = loader_pin_memory
        self.loader_drop_last = loader_drop_last
        self.loader_timeout = loader_timeout

        # Resolve masks and the values they should take given inputs
        self._resolve_dataset_split_indices()

        # Set l and u indices according to mask arguments
        # and need to check that they arent part of
        if labelled_indices is not None and unlabelled_indices is not None:
            if set.intersection(set(labelled_indices), set(unlabelled_indices)):
                raise ValueError("There is overlap between labelled and unlabelled samples")
            self.l_indices = labelled_indices
            self.u_indices = unlabelled_indices
        elif labelled_indices is not None and unlabelled_indices is None:
            self.l_indices = labelled_indices
            updated_u_indices = list(set(self.train_indices) - set(labelled_indices))
            self.unlabelled_indices = updated_u_indices
            self.u_indices = updated_u_indices
        else:
            logger.info("## Labelled and/or unlabelled mask unspecified")
            self.random_label_size = random_label_size
            self.process_random(random_seed)
        self._ensure_no_l_or_u_leaks()
        self._top_unlabelled_set(hit_ratio_at)

    def _ensure_no_split_leaks(self) -> None:
        tt = set.intersection(set(self.train_indices), set(self.test_indices))
        tv, vt = False, False
        if self.validation_indices is not None:
            tv = set.intersection(set(self.train_indices), set(self.validation_indices))
            vt = set.intersection(set(self.validation_indices), set(self.test_indices))
        if tv or tt or vt:
            raise ValueError("There is overlap between the split indices supplied")

    def _ensure_no_empty_train(self) -> None:
        """ensures that the train set is not empty, as there is no need to
        do anything if its empty
        """
        if len(self.train_indices) == 0:
            raise ValueError("The train set is empty")

    def _ensure_no_l_or_u_leaks(self) -> None:
        """ensures that there are no leaks of labelled or unlabelled indices
        in the validation or tests indices
        """
        if self.validation_indices is not None:
            v_overlap = set.intersection(set(self.l_indices), set(self.validation_indices))
            if v_overlap:
                raise ValueError(
                    f"There is {len(v_overlap)} sample overlap between the labelled indices and the validation"
                )
            v_overlap = set.intersection(set(self.u_indices), set(self.validation_indices))
            if v_overlap:
                raise ValueError(
                    f"There is {len(v_overlap)} sample overlap between the unlabelled indices and the validation "
                )

        if self.test_indices is not None:
            t_overlap = set.intersection(set(self.l_indices), set(self.test_indices))
            if t_overlap:
                raise ValueError(
                    f"There is {len(t_overlap)} sample overlap between the labelled indices and the validation"
                )

            # save memory by using same variables
            t_overlap = set.intersection(set(self.u_indices), set(self.test_indices))
            if t_overlap:
                raise ValueError(
                    f"There is {len(t_overlap)} sample overlap between the unlabelled indices and the validation"
                )

    def _resolve_dataset_split_indices(self) -> None:
        """This function is used to resolve what values the indices should be given
        when only a partial subset of them is supplied
        """
        train_flag = self.train_indices is not None
        valid_flag = self.validation_indices is not None
        test_flag = self.test_indices is not None

        # Different cases for presence of training mask, validation mask, test mask
        # TTT Case 0: All masks supplied; check for overlaps to avoid leaks
        if train_flag and valid_flag and test_flag:
            pass

        # TTF Case 1: Any remaining samples become test
        elif train_flag and valid_flag and not test_flag:
            remaining_indices = set(range(len(self.dataset))) - set.union(
                set(self.train_indices), set(self.validation_indices)
            )
            self.test_indices = list(remaining_indices)

        # TFT Case 2: No validation, set any remaining as train (labelled and unlabelled handled seperately)
        elif train_flag and not valid_flag and test_flag:
            remaining_indices = set(range(len(self.dataset))) - set.union(
                set(self.train_indices), set(self.test_indices)
            )
            self.train_indices = set.union(set(self.train_indices), remaining_indices)
            self.train_indices = list(self.train_indices)

        # TFF Case 3: Only train, set others as test
        elif train_flag and not valid_flag and not test_flag:
            remaining_indices = set(range(len(self.dataset))) - set(self.train_indices)
            self.test_indices = list(remaining_indices)

        # FTT Case 4: No train, set any remaining as train
        elif not train_flag and valid_flag and test_flag:
            remaining_indices = set(range(len(self.dataset))) - set.union(
                set(self.validation_indices), set(self.test_indices)
            )
            self.train_indices = list(remaining_indices)

        # FTF Case 5: Only validation, send an error
        elif not train_flag and valid_flag and not test_flag:
            raise ValueError("No train or test specified, too ambigious to set values")

        # FFT Case 6: Only test, send an error
        elif not train_flag and not valid_flag and test_flag:
            raise ValueError("No train or validation specified, too ambigious to set values")

        # FFF Case 7: No masks given, everything is a training observation
        elif not train_flag and not valid_flag and not test_flag:
            self.train_indices = list(range(len(self.dataset)))

        else:
            raise ValueError("Invalid split indices provided")

        self._ensure_no_empty_train()
        self._ensure_no_split_leaks()

    def __len__(self) -> int:
        # Override this if necessary
        return len(self.dataset)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor]:
        # So that one can access samples by index directly
        return self.dataset[idx]

    def _top_unlabelled_set(self, percentage: Optional[Union[int, float]] = None) -> None:
        if percentage is None:
            self.top_unlabelled = None
        else:
            if isinstance(percentage, int):
                percentage /= 100
            assert 0 < percentage < 1, "hit ratio's percentage should be strictly between 0 and 1 (or 0 and 100)"
            ixs = np.array(self.u_indices)
            percentage = int(percentage * len(ixs))
            y = torch.stack(self.get_sample_labels(ixs)).squeeze()
            threshold = np.sort(y.abs())[-percentage]
            self.top_unlabelled = set(ixs[(y.abs() >= threshold).numpy().astype(bool)])

    def get_train_set(self) -> Dataset:
        train_subset = Subset(self.dataset, self.train_indices)
        return train_subset

    def get_validation_set(self) -> Optional[Subset]:
        if self.validation_indices is None:
            return None
        validation_subset = Subset(self.dataset, self.validation_indices)
        return validation_subset

    def get_test_set(self) -> Optional[Subset]:
        if self.test_indices is None:
            return None
        test_subset = Subset(self.dataset, self.test_indices)
        return test_subset

    def get_train_loader(self, full: bool = False) -> DataLoader:
        if full:
            # return full training set with unlabelled included (for strategy evaluation)
            train_loader = self.create_loader(Subset(self.dataset, (self.l_indices + self.u_indices)))
            return train_loader
        else:
            return self.get_labelled_loader()

    def get_validation_loader(self) -> Optional[DataLoader]:
        if self.validation_indices is None:
            return None
        validation_loader = self.create_loader(self.get_validation_set())
        return validation_loader

    def get_test_loader(self) -> Optional[DataLoader]:
        if self.test_indices is None:
            return None
        test_loader = self.create_loader(self.get_test_set())
        return test_loader

    def get_unlabelled_loader(self) -> DataLoader:
        return self.create_loader(Subset(self.dataset, self.u_indices))

    def get_labelled_loader(self) -> DataLoader:
        return self.create_loader(Subset(self.dataset, self.l_indices))

    def process_random(self, seed=0) -> None:
        """Processes the dataset to produce a random subsets of labelled and unlabelled
        samples from the dataset based on the ratio given at initialisation and creates
        the data_loaders
        """
        if isinstance(self.random_label_size, float):
            assert 0 < self.random_label_size < 1, "if a float, random_label_size should be between 0 and 1"
            num_labelled = int(self.random_label_size * len(self.train_indices))
        else:
            num_labelled = self.random_label_size

        logger.info("## Randomly generating labelled subset with {} samples from the train data".format(num_labelled))
        random.seed(seed)
        l_indices = set(random.sample(self.train_indices, num_labelled))
        u_indices = set(self.train_indices) - set(l_indices)

        self.l_indices = list(l_indices)
        self.u_indices = list(u_indices)

    def update_train_labels(self, indices: List[int]) -> None:
        """Updates the L and U sets of the dataset

        Different behaviour based on whether this is done in evaluation mode or real mode.
        The difference is that in evaluation mode the dataset already has the label, so it
        is a matter of making sure the observations are moved from the unlabelled set to the
        labelled set.
        """
        self.l_indices = list(set(self.l_indices + indices))
        self.u_indices = list(set(self.u_indices) - set(indices))

    def percentage_labelled(self) -> float:
        total_len = len(self.l_indices) + len(self.u_indices)
        num_labelled = len(self.l_indices)
        return num_labelled / float(total_len)

    def get_sample(self, ds_index: int) -> Tuple[torch.Tensor]:
        return self[ds_index]

    def get_sample_feature_vector(self, ds_index: int) -> torch.Tensor:
        """To be reviewed for deprecation (for datasets without tensors)"""
        sample = self.get_sample(ds_index)
        sample = sample[0].flatten()
        return sample

    def get_sample_feature_vectors(self, ds_indices: List[int]) -> List[torch.Tensor]:
        """To be reviewed for deprecation (for datasets without tensors)"""
        res = []
        for ds_index in ds_indices:
            res.append(self.get_sample_feature_vector(ds_index))
        return res

    def get_sample_labels(self, ds_indices: List[int]) -> List[torch.Tensor]:
        res = []
        for ds_index in ds_indices:
            res.append(self[ds_index][-1])  # assumes labels are last in output of dataset
        return res

    def create_loader(self, dataset: Dataset) -> DataLoader:
        """Utility to help create dataloader with specifications set at initialisation"""
        batch_size = self.loader_batch_size if isinstance(self.loader_batch_size, int) else len(dataset)
        loader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=self.loader_shuffle,
            sampler=self.loader_sampler,
            batch_sampler=self.loader_batch_sampler,
            num_workers=self.loader_num_workers,
            collate_fn=self.loader_collate_fn,
            pin_memory=self.loader_pin_memory,
            drop_last=self.loader_drop_last,
            timeout=self.loader_timeout,
        )
        return loader

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        """Pretty print a summary of the data_manager contents"""
        str_percentage_labelled = "%.3f" % (self.percentage_labelled())
        str_out = self.__repr__()
        if self.train_indices is not None:
            str_out += "\nTraining set size: {}\n".format(len(self.train_indices))
        if self.labelled_indices is not None:
            str_out += "Labelled: {}, Unlabelled: {}\n".format(len(self.labelled_indices), len(self.unlabelled_indices))
        str_out += "Percentage Labelled: {}".format(str_percentage_labelled)

        return str_out



"""
This module contains methods for scoring samples based on model uncertainty in
classication tasks

This module contains functions for computing the informativeness values
of a given probability distribution (outputs of a model/mc-dropout
prediction, etc.)

"""

import math

import torch


def classification_least_confidence(prob_dist: torch.Tensor, axis: int = -1) -> torch.Tensor:
    r"""Returns the informativeness score of an array using least confidence
    sampling in a 0-1 range where 1 is the most uncertain

    The least confidence uncertainty is the normalised difference between
    the most confident prediction and 100 percent confidence

    Args:
        prob_dist (pytorch tensor): real number tensor whose elements add to 1.0
        sorted (bool): if the probability distribution is pre-sorted from
            largest to smallest
    """
    assert torch.allclose(
        prob_dist.sum(axis), torch.tensor(1.0)
    ), "input should be probability distributions along specified axis"

    simple_least_conf, _ = torch.max(prob_dist, dim=axis)
    num_labels = prob_dist.size(axis)
    normalized_least_conf = (1 - simple_least_conf) * (num_labels / (num_labels - 1))
    return normalized_least_conf


def classification_margin_confidence(prob_dist: torch.Tensor, axis: int = -1) -> torch.Tensor:
    r"""Returns the informativeness score of a probability distribution using
    margin of confidence sampling in a 0-1 range where 1 is the most uncertain
    The margin confidence uncertainty is the difference between the top two
    most confident predictions

    Args:
        prob_dist (pytorch tensor): real number tensor whose elements add to 1.0
        sorted (bool): if the probability distribution is pre-sorted from
            largest to smallest
    """
    assert torch.allclose(
        prob_dist.sum(axis), torch.tensor(1.0)
    ), "input should be probability distributions along specified axis"

    prob_dist, _ = torch.sort(prob_dist, descending=True, dim=axis)
    difference = prob_dist.select(axis, 0) - prob_dist.select(axis, 1)
    margin_conf = 1 - difference
    return margin_conf


def classification_ratio_confidence(prob_dist: torch.Tensor, axis: int = -1) -> torch.Tensor:
    r"""Returns the informativeness score of a probability distribution using
    ratio of confidence sampling in a 0-1 range where 1 is the most uncertain
    The ratio confidence uncertainty is the ratio between the top two most
    confident predictions

    Args:
        prob_dist (pytorch tensor): real number tensor whose elements add to 1.0
        sorted (bool): if the probability distribution is pre-sorted from
            largest to smallest
    """
    assert torch.allclose(
        prob_dist.sum(axis), torch.tensor(1.0)
    ), "input should be probability distributions along specified axis"

    prob_dist, _ = torch.sort(prob_dist, descending=True, dim=axis)  # sort probs so largest is first
    ratio_conf = prob_dist.select(axis, 1) / (prob_dist.select(axis, 0))  # ratio between top two props
    return ratio_conf


def classification_entropy(prob_dist: torch.Tensor, axis: int = -1) -> torch.Tensor:
    r"""Returns the informativeness score of a probability distribution
    using entropy

    The entropy based uncertainty is defined as

    :math:`- \frac{1}{\log(n)} \sum_{i}^{n} p_i \log (p_i)`

    Args:
        prob_dist (pytorch tensor): real number tensor whose elements add to 1.0
    """
    assert torch.allclose(
        prob_dist.sum(axis), torch.tensor(1.0)
    ), "input should be probability distributions along specified axis"

    log_probs = prob_dist * torch.log2(prob_dist)
    raw_entropy = 0 - torch.sum(log_probs, dim=axis)
    normalised_entropy = raw_entropy / math.log2(prob_dist.size(axis))

    return normalised_entropy


def classification_bald(prob_dist: torch.Tensor) -> torch.Tensor:
    """
    Implementation of Bayesian Active Learning by Disagreement (BALD) for classification task

    `reference <https://arxiv.org/pdf/1112.5745.pdf>`__
    :param x: 3D pytorch Tensor of shape n_estimators x n_samples x n_classes
    :return: 1D pytorch tensor of scores
    """

    assert torch.allclose(
        prob_dist.sum(-1), torch.tensor(1.0)
    ), "input should be probability distributions along specified axis"

    return classification_entropy(prob_dist.mean(0), -1) - classification_entropy(prob_dist, -1).mean(0)


def softmax(scores: torch.Tensor, base: float = math.e, axis: int = -1) -> torch.Tensor:
    """Returns softmax array for array of scores

    Converts a set of raw scores from a model (logits) into a
    probability distribution via softmax.

    The probability distribution will be a set of real numbers
    such that each is in the range 0-1.0 and the sum is 1.0.

    Assumes input is a pytorch tensor: tensor([1.0, 4.0, 2.0, 3.0])

    Args
        prediction (pytorch tensor) a pytorch tensor of any positive/negative real numbers.
        base (float) the base for the exponential (default e)
    """
    exps = base ** scores.float()  # exponential for each value in array
    sum_exps = torch.sum(exps, dim=axis, keepdim=True)  # sum of all exponentials
    prob_dist = exps / sum_exps  # normalize exponentials

    return prob_dist



from typing import List

import torch

from pyrelational.data import GenericDataManager
from pyrelational.informativeness import regression_expected_improvement
from pyrelational.models import GenericModel
from pyrelational.strategies.generic_al_strategy import GenericActiveLearningStrategy


class ExpectedImprovementStrategy(GenericActiveLearningStrategy):
    """Implement Expected Improvement Strategy whereby each unlabelled sample is scored based on the
    expected improvement scoring function. The top samples according to this score are selected at each step"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(ExpectedImprovementStrategy, self).__init__(data_manager, model)

    def active_learning_step(self, num_annotate: int) -> List[int]:
        self.model.train(self.l_loader, self.valid_loader)
        output = self.model(self.u_loader)
        max_label = max(self.data_manager.get_sample_labels(self.l_indices))
        uncertainty = regression_expected_improvement(x=output, max_label=max_label)
        ixs = torch.argsort(uncertainty, descending=True).tolist()
        return [self.u_indices[i] for i in ixs[:num_annotate]]



from typing import Dict, Tuple, Type, Union

import pytorch_lightning as pl
import torch
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.utilities.model_helpers import is_overridden
from torch.utils.data import DataLoader

from .generic_model import GenericModel


class LightningModel(GenericModel):
    r"""
    A wrapper for pytorch lightning modules that instantiates and uses a pytorch lightning trainer.

    Example:

    .. code-block:: python

        import torch
        import pytorch_lightning as pl

        class PyLModel(pl.LightningModule):
              def __init__(self, in_dim, out_dim):
                  super(PyLModel, self).()
                  self.linear = torch.nn.Linear(in_dim, out_dim)
           # need to define other train/test steps and optimizers methods required
           # by pytorch-lightning to run this example

        wrapper = LightningModel(
                        PyLModel,
                        model_config={"in_dim":10, "out_dim":1},
                        trainer_config={"epochs":100},
                  )
        wrapper.train(train_loader, valid_loader)
    """

    def __init__(
        self,
        model_class: Type[LightningModule],
        model_config: Union[Dict, str],
        trainer_config: Union[Dict, str],
    ):
        super(LightningModel, self).__init__(model_class, model_config, trainer_config)

    def init_trainer(self) -> Tuple[Trainer, ModelCheckpoint]:
        """

        :param config: dictionary of key:value pairs required to instantiate a trainer object
        :return: a pytorch lightning trainer object
        """
        config = self.trainer_config
        config = _check_pyl_trainer_config(config)
        callbacks = []
        if config["use_early_stopping"]:
            callbacks.append(
                EarlyStopping(
                    monitor=config["monitor_metric_name"],
                    patience=config["patience"],
                    verbose=True,
                    mode=config["monitor_metric_mode"],
                )
            )
        checkpoint_callback = ModelCheckpoint(
            monitor=config["monitor_metric_name"],
            dirpath=config["checkpoints_dir"],
            filename=config["checkpoints_name"],
            save_top_k=config["save_top_k"],
            mode=config["monitor_metric_mode"],
        )
        callbacks.append(checkpoint_callback)

        tracker = pl.loggers.TensorBoardLogger(save_dir=config["checkpoints_dir"], name=config["checkpoints_name"])
        trainer = pl.Trainer(
            callbacks=callbacks,
            logger=tracker,
            gpus=config["gpus"],
            max_epochs=config["epochs"],
            check_val_every_n_epoch=config["period_eval"],
            log_every_n_steps=1,
        )
        return trainer, checkpoint_callback

    def train(self, train_loader: DataLoader, valid_loader: DataLoader = None) -> None:
        trainer, ckpt_callback = self.init_trainer()

        model = self.init_model()
        trainer.fit(model, train_dataloaders=train_loader, val_dataloaders=valid_loader)
        if valid_loader is not None and is_overridden("validation_step", model):
            model.load_state_dict(torch.load(ckpt_callback.best_model_path)["state_dict"])

        self.current_model = model

    def test(self, loader: DataLoader) -> Dict:
        if self.current_model is None:
            raise ValueError("No current model, call 'train(train_loader, valid_loader)' to train the model first")
        trainer, _ = self.init_trainer()
        return trainer.test(self.current_model, dataloaders=loader)[0]

    def __call__(self, loader: DataLoader) -> torch.Tensor:
        """

        :param loader: pytorch dataloader
        :return: model predictions
        """
        if self.current_model is None:
            raise ValueError("No current model, call 'train(train_loader, valid_loader)' to train the model first")
        model = self.current_model
        model.eval()

        with torch.no_grad():
            model_prediction = []
            for x, _ in loader:
                model_prediction.append(model(x).detach().cpu())
        predictions = torch.cat(model_prediction, 0)
        return predictions


def _check_pyl_trainer_config(config: Dict) -> Dict:
    """
    Checks the trainer config for pytorch lightning and adds default values for missing required entries
    :param config: a dictionary with key:values required by the init_trainer function
    :return: dictionary with trainer config
    """
    default = {
        "gpus": 0,
        "epochs": 100,
        "period_eval": 1,
        "checkpoints_dir": "experiment_logs/",
        "checkpoints_name": "run",
        "monitor_metric_name": "loss",
        "monitor_metric_mode": "min",
        "use_early_stopping": False,
        "patience": 100,
        "save_top_k": 1,
    }
    config = {**default, **config}
    return config



from typing import List

import torch

from pyrelational.data import GenericDataManager
from pyrelational.informativeness import regression_upper_confidence_bound
from pyrelational.models import GenericModel
from pyrelational.strategies.generic_al_strategy import GenericActiveLearningStrategy


class UpperConfidenceBoundStrategy(GenericActiveLearningStrategy):
    """Implements Upper Confidence Bound Strategy whereby unlabelled samples are scored and queried based on the
    UCB scorer"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel, kappa: float = 1.0):
        super(UpperConfidenceBoundStrategy, self).__init__(data_manager, model)
        self.kappa = kappa

    def active_learning_step(self, num_annotate: int) -> List[int]:
        self.model.train(self.l_loader, self.valid_loader)
        output = self.model(self.u_loader)
        uncertainty = regression_upper_confidence_bound(x=output, kappa=self.kappa)
        ixs = torch.argsort(uncertainty, descending=True).tolist()
        return [self.u_indices[i] for i in ixs[:num_annotate]]



import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Type, Union

from torch.utils.data import DataLoader


class GenericModel(ABC):
    def __init__(
        self,
        model_class: Type[Any],
        model_config: Union[str, Dict],
        trainer_config: Union[str, Dict],
    ):
        """
        Abstract class used to wrap models to interact with the ActiveLearningStrategy modules.
        It handles model instantiation at each iteration, training, testing, and queries.

        :param model_class: a model constructor (e.g. torch.nn.Linear)
        :param model_config: a dictionary containing the config required to instantiate a model form the model_class
                (e.g. {in_features=100, out_features=34, bias=True, device=None, dtype=None} for a torch.nn.Linear
                constructor)
        :param trainer_config: a dictionary containing the config required to instantiate the trainer module/function
        """
        super(GenericModel, self).__init__()

        self.model_class = model_class
        self.model_config = json.load(open(model_config, "r")) if isinstance(model_config, str) else model_config
        self.current_model = None
        self.trainer_config = (
            json.load(open(trainer_config, "r")) if isinstance(trainer_config, str) else trainer_config
        )

    def init_trainer(self, trainer_config: Dict) -> Any:
        """

        :param trainer_config: a dictionary containing the config required to instantiate the trainer module/function
        :return: trainer module/function
        """
        pass

    def init_model(self) -> Any:
        """

        :return: an instance of self.model_class based on self.model_config
        """
        return self.model_class(**self.model_config)

    @abstractmethod
    def train(self, train_loader: DataLoader, valid_loader: DataLoader = None) -> None:
        """

        :param train_loader: pytorch dataloader for training set
        :param valid_loader: pytorch dataloader for validation set
        :return: none
        """
        pass

    @abstractmethod
    def test(self, loader: DataLoader) -> Dict:
        """

        :param loader: pytorch dataloader for test set
        :return: performance metrics
        """
        pass

    def __call__(self, loader: DataLoader) -> Any:
        """

        :param loader: pytorch dataloader
        :return: uncertainties for each sample in dataloader
        """
        pass

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.model_class.__name__})"



"""
This module contains methods for scoring samples based on distances between
featurization of samples. These scorers are task-agnostic.
"""

import inspect
import logging
from typing import Any, Callable, List, Optional, Union, get_args

import numpy as np
import sklearn.cluster as sklust
import torch
from sklearn.base import ClusterMixin
from sklearn.metrics import pairwise_distances_argmin, pairwise_distances_argmin_min
from torch.utils.data import DataLoader

logging.basicConfig()
logger = logging.getLogger()

Array = Union[torch.Tensor, np.ndarray, List]


def relative_distance(
    query_set: Union[Array, DataLoader],
    reference_set: Union[Array, DataLoader],
    metric: Optional[Union[str, Callable]] = "euclidean",
    axis: int = -1,
) -> torch.Tensor:
    """
    Function that return the minimum distance, according to input metric, from each sample in the query_set to the
    samples in the reference set.

    :param query_set: input containing the features of samples in the queryable pool. query set should either be an
        array-like object or a pytorch dataloader whose first element in each bactch is a featurisation of the samples
        in the batch.
    :param reference_set: input containing the features of samples already queried samples against which the distances
        are computed. reference set should either be an array-like object or a pytorch dataloader whose first element in
        each bactch is a featurisation of the samples in the batch.
    :param metric: defines the metric to be used to compute the distance. This should be supported by scikit-learn
        pairwise_distances function.
    :param axis: integer indicating which dimension the features are
    :return: pytorch tensor of dimension the number of samples in query_set containing the minimum distance from each
        sample to the reference set
    """
    if isinstance(query_set, get_args(Array)):
        query_set = np.array(query_set)
        query_set = query_set.reshape((query_set.shape[0], -1))

    if isinstance(reference_set, get_args(Array)):
        reference_set = np.array(reference_set)
        reference_set = reference_set.reshape((reference_set.shape[0], -1))

    if isinstance(reference_set, np.ndarray) and isinstance(query_set, np.ndarray):

        _, distances = pairwise_distances_argmin_min(query_set, reference_set, metric=metric, axis=axis)
    elif isinstance(reference_set, np.ndarray) and isinstance(query_set, DataLoader):
        distances = []
        for q in query_set:
            q = q[0].reshape((q[0].shape[0], -1))
            distances.append(pairwise_distances_argmin_min(q, reference_set, metric=metric, axis=axis)[1])
        distances = np.hstack(distances)

    elif isinstance(reference_set, DataLoader) and isinstance(query_set, np.ndarray):
        distances = []
        for r in reference_set:
            r = r[0].reshape((r[0].shape[0], -1))
            distances.append(pairwise_distances_argmin_min(query_set, r, metric=metric, axis=axis)[1])
        distances = np.min(np.vstack(distances), axis=0)

    elif isinstance(reference_set, DataLoader) and isinstance(query_set, DataLoader):
        distances = []
        for q in query_set:
            temp = []
            q = q[0].reshape((q[0].shape[0], -1))
            for r in reference_set:
                r = r[0].reshape((r[0].shape[0], -1))
                temp.append(pairwise_distances_argmin_min(q, r, metric=metric, axis=axis)[1])
            distances.append(np.min(np.vstack(temp), axis=0))
        distances = np.hstack(distances)

    else:
        raise TypeError("reference_set and query_set should either be an array_like structure or a pytorch DataLoader")

    return torch.from_numpy(distances).float()


def representative_sampling(
    query_set: Union[Array, DataLoader],
    num_annotate: Optional[int] = None,
    clustering_method: Union[str, ClusterMixin] = "KMeans",
    **clustering_kwargs: Optional[Any],
) -> Array:
    """
    Function that selects representative samples of the query set. Representative selection relies on clustering
    algorithms in scikit-learn.

    :param query_set: input containing the features of samples in the queryable pool. query set should either be an
        array-like object or a pytorch dataloader whose first element in each bactch is a featurisation of the samples
        in the batch
    :param num_annotate: number of representative samples to identify
    :param clustering_method: name, or instantiated class, of the clustering method to use
    :param clustering_kwargs: arguments to be passed to instantiate clustering class if a string is passed to
        clustering_method
    :return: array-like containing the indices of the representative samples identified
    """

    if isinstance(query_set, DataLoader):
        out = []
        for q in query_set:
            out.append(q[0].reshape((q[0].shape[0], -1)))
        query_set = torch.cat(out, 0)
    query_set = np.array(query_set)

    if num_annotate is None and hasattr(clustering_method, "n_clusters"):
        num_annotate = clustering_method.n_clusters

    if num_annotate is not None and (
        num_annotate >= query_set.shape[0]
    ):  # if there are less samples than sought queries, return everything
        return np.arange(query_set.shape[0])

    if isinstance(clustering_method, str) and hasattr(sklust, clustering_method):
        clustering_method = getattr(sklust, clustering_method)
        if "n_clusters" in inspect.getfullargspec(clustering_method).args:
            clustering_kwargs["n_clusters"] = num_annotate
        clustering_cls = clustering_method(**clustering_kwargs)
    elif isinstance(clustering_method, str):
        raise ValueError(f"{clustering_method} is not part of the sklearn package")
    elif isinstance(clustering_method, ClusterMixin):
        clustering_cls = clustering_method
    else:
        raise TypeError(
            """clustering_method argument type not supported, it should be either a string pointing to a method of
            sklearn or an instantiated clustering algorithm subclassing sklearn ClusterMixin"""
        )

    lbls = clustering_cls.fit_predict(query_set)
    if hasattr(clustering_cls, "cluster_centers_indices_"):
        return clustering_cls.cluster_centers_indices_
    elif hasattr(clustering_cls, "cluster_centers_"):
        return get_closest_query_to_centroids(clustering_cls.cluster_centers_, query_set, lbls)
    else:
        logger.warning(
            """Clustering method does not return centroids to identify closest samples,
            returning random sample from each cluster"""
        )
        return get_random_query_from_cluster(lbls)


def get_closest_query_to_centroids(centroids: np.ndarray, query: np.ndarray, cluster_assignment: np.ndarray) -> List:
    out = []
    for i in np.unique(cluster_assignment):
        ixs = np.where(cluster_assignment == i)[0]
        centroid = centroids[i].reshape(1, -1)
        subquery = query[ixs]
        j = pairwise_distances_argmin(centroid, subquery).item()
        out.append(ixs[j])
    return out


def get_random_query_from_cluster(cluster_assignment: np.ndarray) -> List:
    out = []
    for i in np.unique(cluster_assignment):
        ixs = np.where(cluster_assignment == i)[0]
        out.append(np.random.choice(ixs))
    return out



from abc import ABC
from typing import Any, Dict, List, Optional, Type, Union

import numpy as np
import torch
from pytorch_lightning import LightningModule
from pytorch_lightning.utilities.model_helpers import is_overridden
from torch.utils.data import DataLoader

from .generic_model import GenericModel
from .lightning_model import LightningModel


class GenericEnsembleModel(GenericModel, ABC):
    """
    Generic wrapper for ensemble uncertainty estimator
    """

    def __init__(
        self,
        model_class: Type[Any],
        model_config: Union[str, Dict],
        trainer_config: Union[str, Dict],
        n_estimators: int = 10,
    ):
        super(GenericEnsembleModel, self).__init__(model_class, model_config, trainer_config)
        self.n_estimators = n_estimators

    def init_model(self) -> List[Any]:
        """

        :return: list of models
        """
        return [self.model_class(**self.model_config) for _ in range(self.n_estimators)]

    def __call__(self, loader: DataLoader) -> torch.Tensor:
        """

        :param loader: pytorch dataloader
        :return: model predictions
        """
        if self.current_model is None:
            raise ValueError("No current model, call 'train(train_loader, valid_loader)' to train the model first")

        with torch.no_grad():
            predictions = []
            for model in self.current_model:
                model.eval()
                model_prediction = []
                for x, _ in loader:
                    model_prediction.append(model(x).detach().cpu())
                predictions.append(torch.cat(model_prediction, 0))
            predictions = torch.stack(predictions)
        return predictions


class LightningEnsembleModel(GenericEnsembleModel, LightningModel):
    r"""
    Wrapper for ensemble estimator with pytorch lightning trainer


    Example:

    .. code-block:: python

        import torch
        import pytorch_lightning as pl

        class PyLModel(pl.LightningModule):
           def __init__(self, in_dim, out_dim):
               super(PyLModel, self).()
               self.linear = torch.nn.Linear(in_dim, out_dim)
        # need to define other train/test steps and optimizers methods required
        # by pytorch-lightning to run this example

        wrapper = LightningEnsembleModel(
                     PyLModel,
                     model_config={"in_dim":10, "out_dim":1},
                     trainer_config={"epochs":100},
                     n_estimators=10,
               )
        wrapper.train(train_loader, valid_loader)
        predictions = wrapper(loader)
        assert predictions.size(0) == 10



    """

    def __init__(
        self,
        model_class: Type[LightningModule],
        model_config: Union[Dict, str],
        trainer_config: Union[Dict, str],
        n_estimators: int = 10,
    ):
        super(LightningEnsembleModel, self).__init__(
            model_class, model_config, trainer_config, n_estimators=n_estimators
        )

    def train(self, train_loader: DataLoader, valid_loader: Optional[DataLoader] = None) -> None:
        models = self.init_model()
        self.current_model = []
        for model in models:
            trainer, ckpt_callback = self.init_trainer()
            trainer.fit(model, train_dataloaders=train_loader, val_dataloaders=valid_loader)
            if valid_loader is not None and is_overridden("validation_step", model):
                model.load_state_dict(torch.load(ckpt_callback.best_model_path)["state_dict"])
            self.current_model.append(model.cpu())

    def test(self, loader: DataLoader) -> Dict:
        if self.current_model is None:
            raise ValueError("No current model, call 'train(train_loader, valid_loader)' to train the model first")
        trainer, _ = self.init_trainer()
        output = [trainer.test(model, dataloaders=loader)[0] for model in self.current_model]
        # return average score across ensemble
        performances = {}
        for k in output[0].keys():
            performances[k] = np.mean([o[k] for o in output])
        return performances



from pyrelational.data import GenericDataManager
from pyrelational.informativeness import regression_thompson_sampling
from pyrelational.models import GenericModel
from pyrelational.strategies.regression.generic_regression_strategy import (
    GenericRegressionStrategy,
)


class ThompsonSamplingStrategy(GenericRegressionStrategy):
    """Implements Thompson Sampling Strategy whereby unlabelled samples are scored and queried based on the
    thompson sampling scorer"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(ThompsonSamplingStrategy, self).__init__(data_manager, model)
        self.scoring_fn = regression_thompson_sampling



# flake8: noqa
from pyrelational.strategies.task_agnostic.random_acquisition_strategy import (
    RandomAcquisitionStrategy,
)
from pyrelational.strategies.task_agnostic.relative_distance_strategy import (
    RelativeDistanceStrategy,
)
from pyrelational.strategies.task_agnostic.representative_sampling_strategy import (
    RepresentativeSamplingStrategy,
)



"""Representative sampling based active learning strategy
"""
from typing import Any, List, Optional, Union

import torch
from sklearn.base import ClusterMixin

from pyrelational.data import GenericDataManager
from pyrelational.informativeness import representative_sampling
from pyrelational.models import GenericModel
from pyrelational.strategies.generic_al_strategy import GenericActiveLearningStrategy


class RepresentativeSamplingStrategy(GenericActiveLearningStrategy):
    """Representative sampling based active learning strategy"""

    def __init__(
        self,
        data_manager: GenericDataManager,
        model: GenericModel,
        clustering_method: Union[str, ClusterMixin] = "KMeans",
        **clustering_kwargs: Any,
    ):
        super(RepresentativeSamplingStrategy, self).__init__(data_manager, model)
        self.clustering_method = clustering_method
        self.clustering_kwargs = clustering_kwargs

    def active_learning_step(
        self,
        num_annotate: Optional[int] = None,
    ) -> List[int]:
        unlabelled_features = torch.stack(self.data_manager.get_sample_feature_vectors(self.u_indices))
        representative_samples = representative_sampling(
            unlabelled_features,
            num_annotate=num_annotate,
            clustering_method=self.clustering_method,
            **self.clustering_kwargs,
        )
        return [self.u_indices[i] for i in representative_samples]



from typing import List

import torch

from pyrelational.data import GenericDataManager
from pyrelational.informativeness import relative_distance
from pyrelational.models import GenericModel
from pyrelational.strategies.generic_al_strategy import GenericActiveLearningStrategy


class RelativeDistanceStrategy(GenericActiveLearningStrategy):
    """Diversity sampling based active learning strategy"""

    def __init__(
        self,
        data_manager: GenericDataManager,
        model: GenericModel,
    ):
        super(RelativeDistanceStrategy, self).__init__(data_manager, model)

    def active_learning_step(self, num_annotate: int, metric: str = "euclidean") -> List[int]:
        scores = relative_distance(self.u_loader, self.l_loader, metric=metric)
        ixs = torch.argsort(scores, descending=True).tolist()
        return [self.u_indices[i] for i in ixs[:num_annotate]]



from pyrelational.strategies.regression.bald_strategy import (
    BALDStrategy,
    SoftBALDStrategy,
)
from pyrelational.strategies.regression.expected_improvement_strategy import (
    ExpectedImprovementStrategy,
)
from pyrelational.strategies.regression.greedy_strategy import GreedyStrategy
from pyrelational.strategies.regression.least_confidence_strategy import (
    LeastConfidenceStrategy,
)
from pyrelational.strategies.regression.thompson_sampling_strategy import (
    ThompsonSamplingStrategy,
)
from pyrelational.strategies.regression.upper_confidence_bound_strategy import (
    UpperConfidenceBoundStrategy,
)



"""
Active learning using entropy based confidence uncertainty measure
between classes in the posterior predictive distribution to
choose which observations to propose to the oracle
"""

from pyrelational.data import GenericDataManager
from pyrelational.informativeness import classification_entropy
from pyrelational.models import GenericModel
from pyrelational.strategies.classification.generic_classification_strategy import (
    GenericClassificationStrategy,
)


class EntropyClassificationStrategy(GenericClassificationStrategy):
    """Implements Entropy Classification Strategy whereby unlabelled samples are scored and queried based on
    entropy"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(EntropyClassificationStrategy, self).__init__(data_manager, model)
        self.scoring_fn = classification_entropy



from pyrelational.data import GenericDataManager
from pyrelational.informativeness import regression_greedy_score
from pyrelational.models import GenericModel
from pyrelational.strategies.regression.generic_regression_strategy import (
    GenericRegressionStrategy,
)


class GreedyStrategy(GenericRegressionStrategy):
    """Implements Greedy Strategy whereby unlabelled samples are queried based on their predicted mean value
    by the model"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(GreedyStrategy, self).__init__(data_manager, model)
        self.scoring_fn = regression_greedy_score



from pyrelational.datasets.classification import (
    BreastCancerDataset,
    Checkerboard2x2Dataset,
    Checkerboard4x4Dataset,
    CreditCardDataset,
    DigitDataset,
    FashionMNIST,
    GaussianCloudsDataset,
    StriatumDataset,
    SynthClass1,
    SynthClass2,
    SynthClass3,
    UCIGlass,
    UCIParkinsons,
    UCISeeds,
)
from pyrelational.datasets.regression import (
    DiabetesDataset,
    SynthReg1,
    SynthReg2,
    UCIAirfoil,
    UCIConcrete,
    UCIEnergy,
    UCIPower,
    UCIWine,
    UCIYacht,
)
from pyrelational.datasets.uci_datasets import UCIDatasets



"""Regression datasets that can be used for benchmarking AL strategies
"""

import numpy as np
import torch
from sklearn.datasets import load_diabetes, make_regression
from sklearn.model_selection import KFold
from torch.utils.data import Dataset

from .uci_datasets import UCIDatasets


class SynthReg1(Dataset):
    """Synthetic dataset for active learning on a regression based task

    Simple 1 dof regression problem that can be placed into two types
    of AL situations as described in the module docstring

    :param n_splits: an int describing the number of class stratified
            splits to compute
    :param size: an int describing the number of observations the dataset
            is to have
    :param random_seed: random seed for reproducibility on splits
    """

    def __init__(self, n_splits=5, size=1000, random_seed=1234):
        super(SynthReg1, self).__init__()
        self.size = size
        self.random_seed = random_seed
        self.n_splits = n_splits

        X, y = make_regression(n_samples=size, n_features=1, n_targets=1, random_state=random_seed)

        self.x = torch.FloatTensor(X)
        self.y = torch.FloatTensor(y)

        kf = KFold(n_splits=n_splits)
        self.data_splits = kf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class SynthReg2(Dataset):
    """Synthetic dataset for active learning on a regression based task

    A more challenging dataset than SynthReg1 wherein we see a periodic
    pattern with 2 degrees of freedom.

    :param n_splits: an int describing the number of class stratified
            splits to compute
    :param size: an int describing the number of observations the dataset
            is to have
    :param random_seed: random seed for reproducibility on splits
    """

    def __init__(self, n_splits=5, size=1000, random_seed=1234):
        super(SynthReg2, self).__init__()
        self.size = size
        self.random_seed = random_seed
        self.n_splits = n_splits

        # Samples
        zdata = 15 * np.random.random(size)
        xdata = np.sin(zdata) + 0.1 * np.random.randn(size)
        ydata = np.cos(zdata) + 0.1 * np.random.randn(size)

        # Convert
        zdata = torch.FloatTensor(zdata)
        xdata = torch.FloatTensor(xdata)
        ydata = torch.FloatTensor(ydata)

        self.x = torch.vstack([zdata, xdata]).T
        self.y = ydata

        kf = KFold(n_splits=n_splits)
        self.data_splits = kf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class DiabetesDataset(Dataset):
    """A small regression dataset for examples

    From Bradley Efron, Trevor Hastie, Iain Johnstone and
    Robert Tibshirani (2004) “Least Angle Regression,”
    Annals of Statistics (with discussion), 407-499.

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, n_splits=5):
        # Load the diabetes dataset
        diabetes_X, diabetes_y = load_diabetes(return_X_y=True)
        self.x = torch.FloatTensor(diabetes_X)
        self.y = torch.FloatTensor(diabetes_y)

        kf = KFold(n_splits=n_splits)
        self.data_splits = kf.split(self.x, self.y)
        self.data_splits = [(idx[0], idx[1]) for idx in self.data_splits]

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class UCIRegression(Dataset):
    """UCI regression dataset base class

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, name, data_dir="/tmp/", n_splits=5):
        super(UCIRegression, self).__init__()
        dataset = UCIDatasets(name=name, data_dir=data_dir, n_splits=n_splits)
        self.data_dir = dataset.data_dir
        self.name = dataset.name
        self.data_splits = dataset.data_splits

        dataset = dataset.get_simple_dataset()
        self.len_dataset = len(dataset)
        self.x = dataset[:][0]
        self.y = dataset[:][1].squeeze()

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class UCIConcrete(UCIRegression):
    """UCI housing dataset

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIConcrete, self).__init__(name="concrete", data_dir=data_dir, n_splits=n_splits)


class UCIEnergy(UCIRegression):
    """UCI housing dataset

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIEnergy, self).__init__(name="energy", data_dir=data_dir, n_splits=n_splits)


class UCIPower(UCIRegression):
    """UCI housing dataset

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIPower, self).__init__(name="power", data_dir=data_dir, n_splits=n_splits)


class UCIWine(UCIRegression):
    """UCI housing dataset

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIWine, self).__init__(name="wine", data_dir=data_dir, n_splits=n_splits)


class UCIYacht(UCIRegression):
    """UCI housing dataset

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIYacht, self).__init__(name="yacht", data_dir=data_dir, n_splits=n_splits)


class UCIAirfoil(UCIRegression):
    """UCI Airfoil dataset

    :param n_splits: an int describing the number of class stratified
        splits to compute
    """

    def __init__(self, data_dir="/tmp/", n_splits=5):
        super(UCIAirfoil, self).__init__(name="airfoil", data_dir=data_dir, n_splits=n_splits)



"""This module defines the acquisition manager which utilises
the data manager, sampling functions, and model to create acquisition
functions and general arbiters of the active learning pipeline
"""
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
from tabulate import tabulate
from torch.utils.data import DataLoader

from pyrelational.data.data_manager import GenericDataManager
from pyrelational.models.generic_model import GenericModel

logger = logging.getLogger()


class GenericActiveLearningStrategy(ABC):
    """The Acquisition Manager uses the sampling functions and data manager
    to define specific acquisition/active learning strategies and hence
    drive the active learning process

    :param data_manager: an pyrelational data manager
            which keeps track of what has been labelled and creates data loaders for
            active learning
    :param model: An pyrelational model
            which serves as the machine learning model for the data in the
            data manager
    """

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(GenericActiveLearningStrategy, self).__init__()
        self.data_manager = data_manager
        self.model = model
        self.iteration = 0

        # Datastructures for logging values of interest
        self.performances = defaultdict(dict)
        self.labelled_by = {}
        self.log_labelled_by(data_manager.l_indices, tag="Initialisation")

    def theoretical_performance(self, test_loader: Optional[DataLoader] = None) -> Dict:
        """Returns the performance of the full labelled dataset against the
        test data. Typically used for evaluation to establish theoretical benchmark
        of model performance given all available training data is labelled. The
        "horizontal" line in area under learning curve plots for active learning

        Would not make much sense when we are doing active learning for the real
        situation, hence not part of __init__

        :param test_loader: Pytorch Data Loader with
                test data compatible with model, optional as often the test loader can be
                generated from data_manager but is here for case when it hasn't been defined
                or there is a new test set.
        """
        self.model.train(self.train_loader, self.valid_loader)

        # use test loader in data_manager if there is one
        result = self.model.test(self.test_loader if test_loader is None else test_loader)
        if self.data_manager.top_unlabelled is not None:
            result["hit_ratio"] = np.nan
        self.performances["full"] = result
        # make sure that theoretical best model is not stored
        self.model.current_model = None
        return self.performances["full"]

    def current_performance(self, test_loader: Optional[DataLoader] = None, query: Optional[List[int]] = None) -> Dict:
        if self.model.current_model is None:  # no AL steps taken so far
            self.model.train(self.l_loader, self.valid_loader)

        if test_loader is None:
            result = self.model.test(self.test_loader)
        else:
            result = self.model.test(test_loader)
        # reset current model, to avoid issues when model does not need to be trained during al_step.
        self.model.current_model = None

        if self.data_manager.top_unlabelled is not None:
            result["hit_ratio"] = (
                np.nan
                if query is None
                else len(set(query) & self.data_manager.top_unlabelled) / len(self.data_manager.top_unlabelled)
            )
        return result

    @abstractmethod
    def active_learning_step(self, num_annotate: int) -> List[int]:
        """Implements a single step of the active learning strategy stopping and returning the
        unlabelled observations to be labelled as a list of dataset indices

        :param num_annotate: number of observations from u to suggest for labelling
        """
        pass

    def active_learning_update(self, indices: List[int], oracle_interface: object = None, update_tag: str = "") -> None:
        if oracle_interface is None:
            self.update_annotations(indices)
        else:
            oracle_interface.update_dataset(self.data_manager, indices)
            self.update_annotations(indices)

        # Logging
        self.iteration += 1
        logger.info("Length of labelled %s" % (len(self.l_indices)))
        logger.info("Length of unlabelled %s" % (len(self.u_indices)))
        logger.info("Percentage labelled %s" % self.percentage_labelled)
        self.log_labelled_by(indices, tag=update_tag)

    def full_active_learning_run(
        self,
        num_annotate: int,
        num_iterations: Optional[int] = None,
        oracle_interface: object = None,
        test_loader: DataLoader = None,
        return_query_history: bool = False,
    ) -> Optional[Dict]:
        """Given the number of samples to annotate and a test loader
        this method will go through the entire active learning process of training
        the model on the labelled set, and recording the current performance
        based on this. Then it will proceed to compute uncertainties for the
        unlabelled observations, rank them, and get the top num_annotate observations
        labelled to be added to the next iteration's labelled dataset L'. This
        process repeats until there are no observations left in the unlabelled set.

        :param num_annotate: number of observations to get annotated per iteration
        :param num_iterations: number of active learning loop to perform
        :param oracle_interface: undefined for now, this will be entry point for external oracle later
        :param test_loader: test data with which we evaluate the current state of the model given the labelled set L
        :param return_query_history: whether to return the history of queries or not
        :return: optionally returns a dictionary storing the indices of queries at each iteration
        """
        iter_count = 0
        if return_query_history:
            query_history = {}
        while len(self.u_indices) > 0:
            iter_count += 1

            # Obtain samples for labelling and pass to the oracle interface if supplied
            observations_for_labelling = self.active_learning_step(num_annotate)
            if return_query_history:
                query_history[iter_count] = observations_for_labelling

            # Record the current performance
            self.performances[self.iteration] = self.current_performance(
                test_loader=test_loader,
                query=observations_for_labelling,
            )

            self.active_learning_update(
                observations_for_labelling,
                oracle_interface,
                update_tag=str(self.iteration),
            )
            if (num_iterations is not None) and iter_count == num_iterations:
                break

        # Final update the model and check final test performance
        self.model.train(self.l_loader, self.valid_loader)
        self.performances[self.iteration] = self.current_performance(test_loader=test_loader)
        if return_query_history:
            return query_history

    def update_annotations(self, indices: List[int]) -> None:
        self.data_manager.update_train_labels(indices)

    def performance_history(self) -> pd.DataFrame:
        """Constructs a pandas table of performances of the model over the
        active learning iterations
        """
        keys = sorted(set(self.performances.keys()) - {"full"})
        logger.info("KEYS: {}".format(keys))
        df = []
        if "full" in self.performances:
            columns = ["Iteration"] + sorted(list(self.performances["full"].keys()))
            logger.info("COLUMNS: {}".format(columns))
            logger.info("Full inside")
        else:
            if self.iteration in self.performances:
                columns = ["Iteration"] + sorted(list(self.performances[self.iteration].keys()))
            else:
                columns = ["Iteration"]
            logger.info("COLUMNS: {}".format(columns))
            logger.info("Full Missing")
        for k in keys:
            row = [k]
            logger.info(self.performances[k])
            for c in columns[1:]:
                if isinstance(self.performances[k], list):
                    row.append(self.performances[k][0][c])
                else:
                    row.append(self.performances[k][c])
            df.append(row)

        pd_df = pd.DataFrame(df, columns=columns)
        return pd_df

    def log_labelled_by(self, indices: List[int], tag: str = ""):
        """Update the dictionary that records what the observation
        was labelled by. Default behaviour is to map observation to
        iteration at which it was labelled
        """
        for indx in indices:
            self.labelled_by[indx] = tag

    @property
    def u_indices(self) -> List[int]:
        return self.data_manager.u_indices

    @property
    def u_loader(self) -> DataLoader:
        return self.data_manager.get_unlabelled_loader()

    @property
    def l_indices(self) -> List[int]:
        return self.data_manager.l_indices

    @property
    def l_loader(self) -> DataLoader:
        return self.data_manager.get_labelled_loader()

    @property
    def train_loader(self) -> DataLoader:
        return self.data_manager.get_train_loader(full=True)

    @property
    def valid_loader(self) -> DataLoader:
        return self.data_manager.get_validation_loader()

    @property
    def test_loader(self) -> DataLoader:
        return self.data_manager.get_test_loader()

    @property
    def percentage_labelled(self) -> float:
        return self.data_manager.percentage_labelled()

    @property
    def dataset_size(self) -> int:
        return len(self.data_manager)

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        """Pretty prints contents inside ActiveLearningManager based on available attributes"""
        str_dm = str(self.data_manager)
        str_model = str(self.model)
        str_dataset_size = str(self.dataset_size)
        str_percentage_labelled = "%.3f" % self.percentage_labelled

        str_out = self.__repr__()
        str_out += "DataManager: %s \n" % str_dm
        str_out += "Model: %s \n" % str_model
        str_out += "Size of Dataset: %s \n" % str_dataset_size
        str_out += "Percentage of Dataset Labelled for Model: %s \n" % (str_percentage_labelled)
        if "full" in self.performances:
            str_out += "Theoretical performance: %s \n" % str(self.performances["full"])
        str_out += "Performance history \n"
        str_out += tabulate(self.performance_history(), tablefmt="pipe", headers="keys")

        return str_out



"""Utility to create datamanagers corresponding to different AL tasks
"""
import random
from collections import defaultdict

import numpy as np
import torch
from sklearn.metrics import pairwise_distances

from pyrelational.data.data_manager import GenericDataManager


def pick_one_sample_per_class(dataset, train_indices):
    """Utility function to randomly pick one sample per class in the
    training subset of dataset and return their index in the dataset.
    This is used for defining an initial state of the labelled subset
    in the active learning task

    :param train_indices: list or iterable with the indices corresponding
        to the training samples in the dataset
    """
    class2idx = defaultdict(list)
    for idx in train_indices:
        idx_class = int(dataset[idx][1])
        class2idx[idx_class].append(idx)

    class_reps = []
    for idx_class in class2idx.keys():
        random_class_idx = random.choice(class2idx[idx_class])
        class_reps.append(random_class_idx)

    return class_reps


def create_warm_start(dataset, **dm_args):
    """Returns a datamanager with 10% randomly labelled data
    from the train indices. The rest of the observations in the training
    set comprise the unlabelled set of observations. We call this
    initialisation a 'warm start' AL task inspired by
    Konyushkova et al. (2017)

    This can be used both for classification and regression type datasets.

    From Ksenia Konyushkova, Raphael Sznitman, Pascal Fua 'Learning Active
    Learning from Data', NIPS 2017

    :param dataset: A pytorch dataset in the style described
        pyrelational.datasets
    :param dm_args: kwargs for any additional keyword arguments to be passed
        into the initialisation of the datamanager.
    """
    dm = GenericDataManager(dataset, **dm_args)
    return dm


def create_classification_cold_start(dataset, train_indices, test_indices, **dm_args):
    """Returns an AL task for benchmarking classification datasets. The
    AL task will sample an example from each of the classes in the training
    subset of the data.

    Please note the current iteration does not utilise a validation set
    as described in the paper

    :param dataset: A pytorch dataset in the style described
        pyrelational.datasets
    :param train_indices: [int] indices corresponding to observations of dataset
        used for training set
    :param test_indices: [int] indices corresponding to observations of dataset
        used for holdout test set
    :param dm_args: kwargs for any additional keyword arguments to be passed
        into the initialisation of the datamanager.
    """
    labelled_indices = pick_one_sample_per_class(dataset, train_indices)
    dm = GenericDataManager(
        dataset, train_indices=train_indices, test_indices=test_indices, labelled_indices=labelled_indices, **dm_args
    )
    return dm


def create_regression_cold_start(dataset, train_indices, test_indices, **dm_args):
    """Create data manager with 2 labelled data samples, where the data samples
    labelled are the pair that have the largest distance between them

    Please note the current iteration does not utilise a validation set
    as described in the paper

    :param dataset: A pytorch dataset in the style described
        pyrelational.datasets
    :param train_indices: [int] indices corresponding to observations of dataset
        used for training set
    :param test_indices: [int] indices corresponding to observations of dataset
        used for holdout test set
    :param dm_args: kwargs for any additional keyword arguments to be passed
        into the initialisation of the datamanager.
    """
    # Find the two samples within the training subset that have the largest distance between them.
    pair_dists = pairwise_distances(dataset[train_indices][:][0])
    sample1_idx, sample2_idx = np.unravel_index(np.argmax(pair_dists, axis=None), pair_dists.shape)
    sample1_idx = train_indices[sample1_idx]  # map to dataset index from local index
    sample2_idx = train_indices[sample2_idx]

    labelled_indices = [sample1_idx, sample2_idx]
    dm = GenericDataManager(
        dataset,
        train_indices=train_indices,
        test_indices=test_indices,
        labelled_indices=labelled_indices,
        **dm_args,
    )
    return dm



"""
Active learning using marginal confidence uncertainty measure
between classes in the posterior predictive distribution to
choose which observations to propose to the oracle
"""
from pyrelational.data import GenericDataManager
from pyrelational.informativeness import classification_margin_confidence
from pyrelational.models import GenericModel
from pyrelational.strategies.classification.generic_classification_strategy import (
    GenericClassificationStrategy,
)


class MarginalConfidenceStrategy(GenericClassificationStrategy):
    """Implements Marginal Confidence Strategy whereby unlabelled samples are scored and queried based on
    the marginal confidence for classification scorer"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(MarginalConfidenceStrategy, self).__init__(data_manager, model)
        self.scoring_fn = classification_margin_confidence



from pyrelational.data import GenericDataManager
from pyrelational.informativeness import regression_least_confidence
from pyrelational.models import GenericModel
from pyrelational.strategies.regression.generic_regression_strategy import (
    GenericRegressionStrategy,
)


class LeastConfidenceStrategy(GenericRegressionStrategy):
    """Implements Least Confidence Strategy whereby unlabelled samples are queried based on their predicted variance
    by the model"""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(LeastConfidenceStrategy, self).__init__(data_manager, model)
        self.scoring_fn = regression_least_confidence



from setuptools import find_packages, setup

"""
pip install -e .
"""

setup_requires = ["pytest-runner"]
tests_require = ["pytest", "pytest-cov", "mock"]

install_requires = [
    "numpy>=1.20.0",
    "pandas>=1.3.0",
    "pytorch-lightning>=1.5",
    "torch>=1.9.0",
    "scikit-learn>=1.0.2",
    "tabulate>=0.7.0",
    "pyreadr>=0.4.4",
    "xlrd>=2.0.1",
    "openpyxl>=3.0.9",
]


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = {}
with open("pyrelational/version.py") as fp:
    exec(fp.read(), version)

setup(
    name="pyrelational",
    description="Python tool box for quickly implementing active learning strategies",
    author="Relation Therapeutics",
    author_email="software@relationrx.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RelationRx/pyrelational",
    packages=find_packages(),
    version=version["__version__"],
    setup_requires=setup_requires,
    tests_require=tests_require,
    install_requires=install_requires,
    python_requires=">=3.8",
)



import copy
import logging
from abc import ABC
from typing import Dict, Optional, Type, Union

import torch
from pytorch_lightning import LightningModule
from torch.nn.modules import Module
from torch.utils.data import DataLoader

from .generic_model import GenericModel
from .lightning_model import LightningModel

logger = logging.getLogger()


class GenericMCDropoutModel(GenericModel, ABC):
    """
    Generic model wrapper for mcdropout uncertainty estimator
    """

    def __init__(
        self,
        model_class: Type[Module],
        model_config: Union[str, Dict],
        trainer_config: Union[str, Dict],
        n_estimators: int = 10,
        eval_dropout_prob: float = 0.2,
    ):
        super(GenericMCDropoutModel, self).__init__(model_class, model_config, trainer_config)
        _check_mc_dropout_model(model_class, model_config)
        self.n_estimators = n_estimators
        self.eval_dropout_prob = eval_dropout_prob

    def __call__(self, loader: DataLoader) -> torch.Tensor:
        """

        :param loader: pytorch dataloader
        :return: model predictions
        """
        if self.current_model is None:
            raise ValueError("No current model, call 'train(train_loader, valid_loader)' to train the model first")
        predictions = []
        model = self.current_model
        model.eval()

        with torch.no_grad():
            _enable_only_dropout_layers(model, self.eval_dropout_prob)
            for _ in range(self.n_estimators):
                model_prediction = []
                for x, _ in loader:
                    model_prediction.append(model(x).detach().cpu())
                predictions.append(torch.cat(model_prediction, 0))
            predictions = torch.stack(predictions)
        return predictions


class LightningMCDropoutModel(GenericMCDropoutModel, LightningModel):
    r"""
    Wrapper for MC Dropout estimator with pytorch lightning trainer

    Example:

    .. code-block:: python

        import torch
        import pytorch_lightning as pl

        class PyLModel(pl.LightningModule):
              def __init__(self, in_dim, out_dim):
                  super(PyLModel, self).()
                  self.linear = torch.nn.Linear(in_dim, out_dim)
           # need to define other train/test steps and optimizers methods required
           # by pytorch-lightning to run this example

        wrapper = LightningMCDropoutModel(
                        PyLModel,
                        model_config={"in_dim":10, "out_dim":1},
                        trainer_config={"epochs":100},
                        n_estimators=10,
                        eval_dropout_prob=0.2,
                  )
        wrapper.train(train_loader, valid_loader)
        predictions = wrapper(loader)
        assert predictions.size(0) == 10

    """

    def __init__(
        self,
        model_class: Type[LightningModule],
        model_config: Union[Dict, str],
        trainer_config: Union[Dict, str],
        n_estimators: int = 10,
        eval_dropout_prob: float = 0.2,
    ):
        super(LightningMCDropoutModel, self).__init__(
            model_class,
            model_config,
            trainer_config,
            n_estimators=n_estimators,
            eval_dropout_prob=eval_dropout_prob,
        )


def _enable_only_dropout_layers(model: Module, p: Optional[float] = None) -> None:
    def enable_dropout_on_module(m):
        if m.__class__.__name__.startswith("Dropout"):
            if isinstance(p, float) and (0 <= p <= 1):
                m.p = p
            elif isinstance(p, float) and (p < 0 or p > 1):
                logger.warning(f"Evaluation dropout probability should be a float between 0 and 1, got {p}")
            m.train()

    model.apply(enable_dropout_on_module)


def _check_mc_dropout_model(model_class: Type[Module], model_config: Dict) -> None:
    model = model_class(**model_config)

    def has_dropout_module(model):
        is_dropout = []
        for m in model.children():
            if m.__class__.__name__.startswith("Dropout"):
                is_dropout.append(True)
            else:
                is_dropout += has_dropout_module(m)
        return is_dropout

    if not any(has_dropout_module(model)):
        raise ValueError("Model provided do not contain any torch.nn.Dropout modules, cannot apply MC Dropout")



from typing import List

import numpy as np
import torch

from pyrelational.data import GenericDataManager
from pyrelational.informativeness import regression_bald
from pyrelational.models import GenericModel
from pyrelational.strategies.regression.generic_regression_strategy import (
    GenericRegressionStrategy,
)


class BALDStrategy(GenericRegressionStrategy):
    """Implements BALD Strategy whereby unlabelled samples are queried based on mutual information score based on
    multiple estimator models."""

    def __init__(self, data_manager: GenericDataManager, model: GenericModel):
        super(BALDStrategy, self).__init__(data_manager, model)
        self.scoring_fn = regression_bald


class SoftBALDStrategy(BALDStrategy):
    """Implements soft BALD Strategy whereby unlabelled samples are queried based on mutual information score based on
    multiple estimator models. In contrast to Bald the query is drawn from unlabelled pool based on probabilities
    derived from bald scores instead of using an argmax operation"""

    def __init__(
        self,
        data_manager: GenericDataManager,
        model: GenericModel,
        temperature: float = 0.5,
    ):
        super(SoftBALDStrategy, self).__init__(data_manager, model)
        assert temperature > 0, "temperature parameter should be greater than 0"
        self.T = torch.tensor(temperature)

    def active_learning_step(self, num_annotate: int) -> List[int]:
        self.model.train(self.l_loader, self.valid_loader)
        output = self.model(self.u_loader)
        scores = self.scoring_fn(x=output) / self.T
        scores = torch.softmax(scores, -1).numpy()
        num_annotate = min(num_annotate, len(self.u_indices))
        return np.random.choice(self.u_indices, size=num_annotate, replace=False, p=scores).tolist()
